/*
This file is part of video3x.

video3x is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

video3x is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with video3x.  If not, see <http://www.gnu.org/licenses/>.

Copyright 2011 Vadim Druzhin <cdslow@mail.ru>
*/

#include <string.h>
#include "vi3x_modules.h"

static char const *Tag="reil";

/* Instance local data */
typedef struct m_priv_s
    {
    int il;
    } m_priv_t;

static int m_init(vi3x_module_t *m);
static int m_fini(vi3x_module_t *m);
static int m_image(vi3x_module_t *m, vi3x_image_t *image);
static void m_help(vi3x_help_t *h);
static void help_il(vi3x_arg_t const *arg, vi3x_help_t const *h);

/* Module arguments */
static vi3x_arg_t ArgMap[]=
    {
    { "il", 0, offsetof(m_priv_t, il), help_il },
    { NULL, 0, 0, NULL }
    };

/* Entry point */
VI3X_MAIN()
    {
    vi3x_reg_filter(Tag, m_init, m_help);
    }

/* Module help */
static void m_help(vi3x_help_t *h)
    {
    h->help_module(Tag, VI3X_VER_STR, "Vadim Druzhin",
        "Change the chroma planes upsampling method",
        "The function of this filter is the same as downsample chroma planes\n"
        "and then upsample them using given interlacing mode.",
        ArgMap
        );
    }

static void help_il(vi3x_arg_t const *arg, vi3x_help_t const *h)
    {
    h->help_param_bool(arg->name,
        "New interlacing mode of chroma planes", "",
        arg->def,
        "- convert to progressive;",
        "- convert to interlaced."
        );
    }

/* Instance initialization */
static int m_init(vi3x_module_t *m)
    {
    /* Verify frame parameters */
    if(m->h < 8)
        {
        m->vi3x.err("[%s] Error: Frame height too small: h=%d < 8 !\n", Tag, m->h);
        return 0;
        }
    if(m->h % 4 != 0)
        {
        m->vi3x.err("[%s] Error: Frame height isn't multiple of 4: h=%d !\n",
            Tag, m->h);
        return 0;
        }

    /* Allocate instance memory */
    m->priv=(m_priv_t *)m->vi3x.alloc(sizeof(*m->priv), Tag);

    /* Parse and verify arguments */
    if(!m->vi3x.parse_args(m->priv, m->args, ArgMap, Tag))
        {
        m->vi3x.free(m->priv);
        return 0;
        }

    /* Publish module interface */
    m->fini=m_fini;
    m->process_image=m_image;

    return 1;
    }

/* Instance finalization */
static int m_fini(vi3x_module_t *m)
    {
    /* Free instance memory */
    m->vi3x.free(m->priv);

    return 1;
    }

/* Frame processor */
static int m_image(vi3x_module_t *m, vi3x_image_t *image)
    {
    if(image->il != m->priv->il)
        {
        if(m->priv->il)
            m->vi3x.frame_to_fields(image);
        else
            m->vi3x.fields_to_frame(image);
        }

    return m->out_image(m, image);
    }

