/*
This file is part of video3x.

video3x is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

video3x is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with video3x.  If not, see <http://www.gnu.org/licenses/>.

Copyright 2011 Vadim Druzhin <cdslow@mail.ru>
*/

#include <stddef.h>
#include <string.h>
#include <stdint.h>
#include <stdarg.h>
#include <limits.h>
#include "vi3x_modules.h"

static char const *Tag="tooncine";
static void (*vi3x_copy_image)(vi3x_image_t *dst, vi3x_image_t const *src);
static void (*vi3x_info)(char const *fmt, ...);
static void (*vi3x_info_va)(char const *fmt, va_list ap);

#define CSIGN_THRESHOLD -4
#define NOISE_SCALE 16

typedef struct frames_s
    {
    vi3x_image_t img;
    int ill;
    } frame_t;

#define G_SIZE 5
#define CINE_DIV 4
#define MIN_DEPTH 2

#define ASS_STR(s) #s
#define AS_STR(s) ASS_STR(s)

/* Instance local data */
typedef struct m_priv_s
    {
    int depth;
    int cine_hi;
    int cine_lo;
    int cine_huge;
    int cine_depth;
    int ill_sens;
    int noise;
    int debug;
    frame_t *frames;
    int il_warned;
    int count;
    int cine;
    int cine_changed;
    int pre_cine;
    int delta;
    int prev_ill;
    int il_max;
    int frame_o;
    } m_priv_t;

static int init_tc(vi3x_module_t *m);
static int fini_tc(vi3x_module_t *m);
static int image_tc(vi3x_module_t *m, vi3x_image_t *image);
static int flush_tc(vi3x_module_t *m, vi3x_image_t *image);

static void reset_priv(m_priv_t *priv);
static void verify_cine(m_priv_t *p, int from);
static void search_cine(m_priv_t *p);
static void force_cine(m_priv_t *p);
static int flush_frames(vi3x_module_t *m);
static int flush_tail(vi3x_module_t *m);
static int count_il(
    vi3x_image_t const *img_bottom, vi3x_image_t const *img_top, int dmin);
static void merge_fields(
    vi3x_image_t *dst, vi3x_image_t const *src, int bottom);
static void dbg_group(m_priv_t const *p,
    int from, int to,
    int scale,
    int i1, char c1_l, char c1_r,
    int i2, char c2_l, char c2_r
    );
static void dbg_msg(m_priv_t const *p, char const *fmt, ...);
static int flush_group(vi3x_module_t *m, int from);
static int group_base(m_priv_t *p, int from);
static int group_ivtc(m_priv_t *p, int cine, int base_ill);
static int group_fix_tail(m_priv_t *p, int cine, int to, int base_ill);
static void group_fix_head(m_priv_t *p, int from, int cine, int base_ill);
static int group_pre_skip(m_priv_t *p, int from);
static int group_pre_cine(vi3x_module_t *m, int from, int i_skip);
static int flush_range(vi3x_module_t *m, int from, int to, int skip);
static int cond_replace_frame(m_priv_t *p, int idst, int isrc, int base_ill);
static void m_help(vi3x_help_t *h);
static void help_depth(vi3x_arg_t const *arg, vi3x_help_t const *h);
static void help_cine_hi(vi3x_arg_t const *arg, vi3x_help_t const *h);
static void help_cine_lo(vi3x_arg_t const *arg, vi3x_help_t const *h);
static void help_cine_huge(vi3x_arg_t const *arg, vi3x_help_t const *h);
static void help_cine_depth(vi3x_arg_t const *arg, vi3x_help_t const *h);
static void help_ill_sens(vi3x_arg_t const *arg, vi3x_help_t const *h);
static void help_noise(vi3x_arg_t const *arg, vi3x_help_t const *h);
static void help_debug(vi3x_arg_t const *arg, vi3x_help_t const *h);

/* Module arguments */
static vi3x_arg_t ArgMap[]=
    {
    { "depth",      30, offsetof(m_priv_t, depth), help_depth },
    { "cine_hi",    12, offsetof(m_priv_t, cine_hi), help_cine_hi },
    { "cine_lo",     6, offsetof(m_priv_t, cine_lo), help_cine_lo },
    { "cine_huge", 100, offsetof(m_priv_t, cine_huge), help_cine_huge },
    { "cine_depth",  4, offsetof(m_priv_t, cine_depth), help_cine_depth },
    { "ill_sens",    4, offsetof(m_priv_t, ill_sens), help_ill_sens },
    { "noise",      48, offsetof(m_priv_t, noise), help_noise },
    { "debug",       0, offsetof(m_priv_t, debug), help_debug },
    { NULL, 0, 0, NULL }
    };

/* Entry point */
VI3X_MAIN()
    {
    vi3x_reg_filter(Tag, init_tc, m_help);
    }

/* Module help */
static void m_help(vi3x_help_t *h)
    {
    h->help_module(Tag, VI3X_VER_STR, "Vadim Druzhin",
        "Cartoon-oriented inverse telecine filter",
        "",
        ArgMap
        );
    }

static void help_depth(vi3x_arg_t const *arg, vi3x_help_t const *h)
    {
    h->help_param_range(arg->name,
        "Maximum number of buffered telecine groups",
        "Each telecine group consist of " AS_STR(G_SIZE) " frames.\n"
        "Small values may decrease an accuracy for low motion scenes,\n"
        "large values may cause the wrong telecine detection for short scenes.",
        arg->def, MIN_DEPTH, INT_MAX
        );
    }

static void help_cine_hi(vi3x_arg_t const *arg, vi3x_help_t const *h)
    {
    h->help_param_range(arg->name,
        "High interlacing coefficient",
        "The interlacing ratio of a highly interlaced frame\n"
        "to the average interlacing of a telecine group scaled by " AS_STR(CINE_DIV) ".\n"
        "This value must be greater than cine_lo.",
        arg->def, 0, INT_MAX
        );
    }

static void help_cine_lo(vi3x_arg_t const *arg, vi3x_help_t const *h)
    {
    h->help_param_range(arg->name,
        "Low interlacing coefficient",
        "The interlacing ratio of an interlaced frame\n"
        "to the average interlacing of a telecine group scaled by " AS_STR(CINE_DIV) ".\n"
        "This value must be lower than cine_hi.",
        arg->def, 0, INT_MAX
        );
    }

static void help_cine_huge(vi3x_arg_t const *arg, vi3x_help_t const *h)
    {
    h->help_param_range(arg->name,
        "Interlacing verification coefficient",
        "If the interlacing ratio of a recombined frame\n"
        "to the average interlacing of a telecine group scaled by " AS_STR(CINE_DIV) "\n"
        "exceeds this value, it will be considered as a telecine mismatch.",
        arg->def, 0, INT_MAX
        );
    }

static void help_cine_depth(vi3x_arg_t const *arg, vi3x_help_t const *h)
    {
    h->help_param_range(arg->name,
        "Minimum number of groups for weak telecine detection", "",
        arg->def, 0, INT_MAX
        );
    }

static void help_ill_sens(vi3x_arg_t const *arg, vi3x_help_t const *h)
    {
    h->help_param_range(arg->name,
        "Interlacing level ratio for a new scene detection", "",
        arg->def, 0, INT_MAX
        );
    }

static void help_noise(vi3x_arg_t const *arg, vi3x_help_t const *h)
    {
    h->help_param_range(arg->name,
        "Noise threshold for the interlacing detector", "",
        arg->def, 0, VI3X_MAX * VI3X_MAX * 3 / NOISE_SCALE
        );
    }

static void help_debug(vi3x_arg_t const *arg, vi3x_help_t const *h)
    {
    h->help_param_bool(arg->name,
        "Enable debug information", "", arg->def, "", "");
    }

/* Instance initialization */
static int init_tc(vi3x_module_t *m)
    {
    m_priv_t tmp_priv;
    vi3x_image_t tmp_image;
    uint8_t *buf;
    int i;

    /* Check frame parameters */
    if(m->h < 4)
        {
        m->vi3x.warn("[%s] Error: Frame height too small: h=%d < 4 !\n", Tag, m->h);
        return 0;
        }
    if(m->h & 1)
        {
        m->vi3x.err("[%s] Error: Uneven frame height: h=%d !\n", Tag, m->h);
        return 0;
        }

    /* Parse and verify arguments */
    if(!m->vi3x.parse_args(&tmp_priv, m->args, ArgMap, Tag))
        return 0;
    if(tmp_priv.depth < tmp_priv.cine_depth + 1)
        tmp_priv.depth=tmp_priv.cine_depth + 1;
    if(tmp_priv.depth < MIN_DEPTH)
        {
        m->vi3x.err("[%s] Error: Depth too small: depth=%d < %d !\n",
            Tag, tmp_priv.depth, MIN_DEPTH);
        return 0;
        }
    tmp_priv.depth*=G_SIZE;

    /* Calculate frame size */
    m->vi3x.init_image(&tmp_image, m->w, m->h, NULL);

    /* Allocate instance memory */
    buf=m->vi3x.alloc(sizeof(*m->priv)
        + (sizeof(*m->priv->frames) + tmp_image.img_sz) * tmp_priv.depth, Tag);

    /* Store arguments */
    m->priv=(m_priv_t *)buf;
    buf+=sizeof(*m->priv);
    *m->priv=tmp_priv;

    /* Initialize image buffers */
    m->priv->frames=(frame_t *)buf;
    buf+=sizeof(*m->priv->frames) * m->priv->depth;
    for(i=0; i < m->priv->depth; ++i)
        {
        buf=m->vi3x.init_image(&m->priv->frames[i].img, m->w, m->h, buf);
        m->priv->frames[i].ill=0;
        }

    /* Initialize state */
    reset_priv(m->priv);
    m->priv->il_warned=0;

    /* Publish module interface */
    m->fps=m->fps * (G_SIZE - 1);
    if(0 == m->fps % G_SIZE)
        m->fps=m->fps / G_SIZE;
    else
        m->fps_div=m->fps_div * G_SIZE;
    m->fini=fini_tc;
    m->process_image=image_tc;
    m->flush=flush_tc;

    /* Remember vi3x functions */
    vi3x_copy_image=m->vi3x.copy_image;
    vi3x_info=m->vi3x.info;
    vi3x_info_va=m->vi3x.info_va;

    return 1;
    }

static void reset_priv(m_priv_t *priv)
    {
    priv->count=0;
    priv->cine=-1;
    priv->cine_changed=0;
    priv->pre_cine=0;
    priv->delta=0;
    priv->prev_ill=0;
    priv->il_max=-1;
    priv->frame_o=0;
    }

static void reset_cine(m_priv_t *priv)
    {
    priv->prev_ill=0;
    priv->cine=-1;
    priv->il_max=-1;
    priv->pre_cine=0;
    }

/* Instance finalization */
static int fini_tc(vi3x_module_t *m)
    {
    /* Free instance memory */
    m->vi3x.free(m->priv);

    return 1;
    }

/* Frame processor */
static int image_tc(vi3x_module_t *m, vi3x_image_t *image)
    {
    m_priv_t *p=m->priv;

    /* Verify frame type */
    if(!image->il && !p->il_warned)
        {
        m->vi3x.warn("[%s] Warning: Input frame marked as progressive, "
            "processing may be inaccurate!\n", Tag);
        p->il_warned=1;
        }

    /* Store image */
    m->vi3x.copy_image(&p->frames[p->count].img, image);
    /* Calculate interlacing levels */
    p->frames[p->count].ill=
        count_il(&p->frames[p->count].img, &p->frames[p->count].img, p->noise);
    ++p->count;

    if(p->count - p->pre_cine >= G_SIZE)
        {
        /* Verify current cine */
        if(p->cine >= 0)
            verify_cine(p, p->pre_cine);

        /* Search new cine */
        if(p->cine < 0 && p->count >= MIN_DEPTH * G_SIZE)
            search_cine(p);

        /* Queue full */
        if(p->cine < 0 && p->count == p->depth)
            {
            if(p->debug)
                dbg_msg(p, "Buffer is full (count=%d)\n", p->count);
            force_cine(p);
            }

        /* Output frames if telecine found */
        if(p->cine >= 0 && p->count - p->pre_cine >= G_SIZE)
            return flush_frames(m);
        }

    return 1;
    }

/* End of processing */
static int flush_tc(vi3x_module_t *m, vi3x_image_t *image)
    {
    int ret=1;

    (void)image; /* Unused */

    while(ret && m->priv->count >= G_SIZE)
        {
        if(m->priv->cine < 0)
            force_cine(m->priv);
        if(m->priv->count - m->priv->pre_cine < G_SIZE)
            break;
        ret=flush_frames(m);
        }

    if(ret && m->priv->count > 0)
        ret=flush_tail(m);

    reset_priv(m->priv);

    return ret;
    }

/* Verify cine in group */
static void verify_cine(m_priv_t *p, int from)
    {
    int il_max, il_max2;
    int i;
    int base_ill;
    int merge_ill;
    int cine;

    /* Calculate basic interlacing level of regular frames */
    base_ill=group_base(p, from);
    cine=from + p->cine;

    /* If interlacing level changed significantly */
    if(p->prev_ill > 0 && base_ill > p->prev_ill * p->ill_sens)
        {
        /* Assume scene change */
        if(p->debug)
            {
            dbg_msg(p, "[%03d]: CINE MISSED: base_ill=%d prev_ill=%d(%d)\n",
                from, base_ill, p->prev_ill, base_ill / p->prev_ill);
            dbg_group(p, from, from + G_SIZE, base_ill, cine, '[', ']', -1, 0, 0);
            }
        reset_cine(p);
        return;
        }

    /* Search interlacing maximum */
    il_max=from + G_SIZE - 1;
    for(i=il_max - 1; i >= from; --i)
        {
        if(p->frames[il_max].ill < p->frames[i].ill)
            il_max=i;
        }

    /* If maximum interlacing don't match cine position */
    if(il_max != cine && il_max != cine + 1
        && p->frames[il_max].ill * CINE_DIV > base_ill * p->cine_lo)
        {
        if(p->debug)
            {
            dbg_msg(p, "[%03d]: CINE MISSED: base_ill=%d ill[%d]=%d(%d) cine=%d\n",
                from, base_ill, il_max, p->frames[il_max].ill,
                p->frames[il_max].ill * CINE_DIV / (base_ill ? base_ill : 1),
                p->cine);
            dbg_group(p, from, from + G_SIZE, base_ill, il_max, '<', '>',
                cine, '[', ']');
            }
        reset_cine(p);
        return;
        }

    /* Select second frame in cine group around il_max */
    il_max2=il_max + 1;
    if(il_max2 >= from + G_SIZE)
        il_max2-=G_SIZE;
    i=il_max - 1;
    if(i < from)
        i+=G_SIZE;
    if(p->frames[il_max2].ill < p->frames[i].ill)
        il_max2=i;

    /* If maximum interlacing don't match cine position */
    if(il_max2 != cine && il_max2 != cine+1
        && p->frames[il_max2].ill * CINE_DIV > base_ill * p->cine_lo)
        {
        if(p->debug)
            {
            dbg_msg(p, "[%03d]: CINE MISSED: base_ill=%d ill[%d]=%d(%d) cine=%d\n",
                from, base_ill, il_max2, p->frames[il_max2].ill,
                p->frames[il_max2].ill * CINE_DIV / (base_ill ? base_ill : 1),
                p->cine);
            dbg_group(p, from, from + G_SIZE, base_ill, il_max2, '<', '>',
                cine, '[', ']');
            }
        reset_cine(p);
        return;
        }

    /* Calculate interlacing of combined frame */
    merge_ill=count_il(&p->frames[cine].img, &p->frames[cine+1].img, p->noise);

    /* Check interlacing level of IVTC frame */
    if(base_ill > 0 && merge_ill * CINE_DIV > base_ill * p->cine_hi)
        {
        /* Wrong telecine structure */
        if(p->debug)
            {
            dbg_msg(p, "[%03d]: CINE MISSED: base_ill=%d merge_ill=%d(%d)\n",
                from, base_ill, merge_ill,
                merge_ill * CINE_DIV / (base_ill ? base_ill : 1));
            dbg_group(p, from, from + G_SIZE, base_ill, il_max, '<', '>',
                cine, '[', ']');
            }
        reset_cine(p);
        return;
        }
    }

/* Search cine position in last group */
static void search_cine(m_priv_t *p)
    {
    int i;
    int il_max, il_max2;
    int il_cine;
    int base_ill;
    int sum_ill;
    int cnl1, cnl2;

    /* Search interlacing maximum */
    il_max=p->count - G_SIZE;
    for(i=il_max + 1; i < p->count; ++i)
        {
        if(p->frames[il_max].ill < p->frames[i].ill)
            il_max=i;
        }

    if(0 == il_max || p->count - 1 == il_max)
        /* Can't detect telecine by edge frames */
        return;

    /* Guess telecine position */
    if(p->frames[il_max-1].ill > p->frames[il_max+1].ill)
        {
        il_max2=il_max - 1;
        il_cine=il_max2;
        }
    else
        {
        il_max2=il_max + 1;
        il_cine=il_max;
        }

    if(0 == il_cine || p->count - 2 == il_cine)
        /* Can't detect telecine by edge frames */
        return;

    /* Calculate basic interlacing level of regular frames */
    i=p->count - G_SIZE;
    if(i == il_cine)
        i=il_cine - 1;
    base_ill=group_base(p, i);

    /* If Interlacing too low */
    if(p->frames[il_max].ill * CINE_DIV < base_ill * p->cine_hi)
        {
        /* Calculate overall interlacing */
        sum_ill=base_ill;
        for(i=il_max; i > G_SIZE; i-=G_SIZE)
            sum_ill+=p->frames[i].ill - base_ill;
        if(sum_ill * CINE_DIV < base_ill * p->cine_hi)
            return;
        if(p->debug)
            {
            dbg_msg(p, "[%03d]: WEAK CINE: base_ill=%d sum_ill[%d]=%d(%d)\n",
                p->count - G_SIZE, base_ill, il_max, sum_ill,
                sum_ill * CINE_DIV / (base_ill ? base_ill : 1),
                p->cine, p->pre_cine);
            dbg_group(p, p->count-G_SIZE, p->count, base_ill,
                il_cine, '[', ']', il_max, '<', '>');
            }
        }

    if(-1 == p->il_max)
        p->il_max=il_max;

    /* If second frame interlacing too low */
    if(p->frames[il_max2].ill * CINE_DIV < base_ill * p->cine_lo)
        {
        if(p->il_max % G_SIZE != il_max % G_SIZE) /* If maximum changed */
            il_max=p->il_max; /* Use previous groups */
        else if(p->count < (p->cine_depth + 1) * G_SIZE) /* If not enough data */
            {
            /* Remember maximum */
            p->il_max=il_max;
            return;
            }
        /* Calculate interlacing of combined frames */
        cnl1=cnl2=0;
        for(i=il_max; i > G_SIZE; i-=G_SIZE)
            {
            cnl1+=count_il(&p->frames[i-1].img, &p->frames[i].img, p->noise);
            cnl2+=count_il(&p->frames[i].img, &p->frames[i+1].img, p->noise);
            }
        if(cnl1<cnl2)
            il_cine=il_max - 1;
        else
            il_cine=il_max;
        }

    /* Got it */
    p->cine=il_cine % G_SIZE;
    p->cine_changed=1;
    p->pre_cine=0;
    /* Shift cine from group edge */
    if(G_SIZE - 1 == p->cine)
        {
        p->pre_cine=p->cine - 1;
        p->cine-=p->pre_cine;
        }
    if(p->debug)
        {
        dbg_msg(p, "[%03d]: CINE FOUND: "
            "base_ill=%d ill[%d]=%d(%d) cine=%d pre_cine=%d\n",
            p->count - G_SIZE, base_ill, il_max, p->frames[il_max].ill,
            p->frames[il_max].ill * CINE_DIV / (base_ill ? base_ill : 1),
            p->cine, p->pre_cine);
        for(i=p->count % G_SIZE; i <= p->count - G_SIZE; i+=G_SIZE)
            dbg_group(p, i, i + G_SIZE, base_ill,
                il_cine, '[', ']', il_max, '<', '>');
        dbg_msg(p, "---\n");
        }
    }

/* Search cine position by all buffered frames */
static void force_cine(m_priv_t *p)
    {
    int i;
    int ill[G_SIZE];
    int n;
    int il_max;

    memset(ill, 0, sizeof(ill));

    /* Calculate summary interlacing */
    if(p->count < G_SIZE*2)
        n=0; /* Take all frames into account */
    else
        n=G_SIZE; /* Skip first group */
    for(i=n; i < p->count; ++i)
        ill[i%G_SIZE]+=p->frames[i].ill;
    n=(p->count - n) / G_SIZE;
    for(i=0; i < p->count % G_SIZE; ++i)
        ill[i]/=n + 1;
    if(n>0)
        {
        for(; i < G_SIZE; ++i)
            ill[i]/=n;
        }

    /* Search maximum */
    il_max=0;
    for(i=1; i < G_SIZE; ++i)
        {
        if(ill[il_max] < ill[i])
            il_max=i;
        }

    /* Select cine position */
    if(ill[(il_max+G_SIZE-1)%G_SIZE] > ill[(il_max+1)%G_SIZE])
        p->cine=(il_max + G_SIZE - 1) % G_SIZE;
    else
        p->cine=il_max;
    p->cine_changed=1;
    p->pre_cine=0;

    /* Shift cine position from group edge */
    if(G_SIZE - 1 == p->cine)
        {
        p->pre_cine=p->cine - 1;
        p->cine-=p->pre_cine;
        }

    if(p->debug)
        {
        int base_ill=0;

        for(i=0; i < G_SIZE; ++i)
            {
            if(i != p->cine && i != (p->cine + 1) % G_SIZE)
                base_ill+=ill[i];
            }
        base_ill=base_ill / (G_SIZE - 2);

        dbg_msg(p, "CINE FORCED: base_ill=%d ill[%d]=%d(%d) cine=%d pre_cine=%d\n",
            base_ill, il_max, ill[il_max],
            ill[il_max] * CINE_DIV / (base_ill ? base_ill : 1),
            p->cine, p->pre_cine);
        }
    }

static int flush_frames(vi3x_module_t *m)
    {
    m_priv_t *p=m->priv;
    int i, j;

    /* Output frames by group */
    i=p->pre_cine;
    for(;;)
        {
        if(!flush_group(m, i))
            return 0;
        i+=G_SIZE;
        if(p->cine < 0)
            break;
        if(p->count - i < G_SIZE)
            break;
        verify_cine(p, i);
        if(p->cine < 0)
            break;
        }

    /* Move leftover frames to begining of the buffer */
    for(j=0; i < p->count; ++j, ++i)
        {
        vi3x_copy_image(&p->frames[j].img, &p->frames[i].img);
        p->frames[j].ill=p->frames[i].ill;
        }
    p->count=j;

    return 1;
    }

/* Output next group of frames */
static int flush_group(vi3x_module_t *m, int from)
    {
    m_priv_t *p=m->priv;
    int cine;
    int base_ill;
    int ivtc;
    char cl, cr;
    int i_skip;

    cine=from + p->cine;

    /* Calculate basic interlacing level of regular frames */
    base_ill=group_base(p, from);
    /* Remember it */
    p->prev_ill=base_ill;

    /* Do IVTC */
    ivtc=group_ivtc(p, cine, base_ill);

    /* Process sequence start */
    if(p->cine_changed)
        {
        /* Search from to skip before group */
        i_skip=group_pre_skip(p, from);

        /* Replace hardly interlaced frames at sequence start */
        if(cine > from - p->pre_cine)
            group_fix_head(p, from - p->pre_cine, cine, base_ill);

        /* Output pre_cine frames */
        if(p->pre_cine > 0)
            {
            if(!group_pre_cine(m, from, i_skip))
                return 0;
            p->pre_cine=0;
            }

        p->cine_changed=0;
        }

    /* Replace hardly interlaced frames after cine position */
    if(group_fix_tail(p, cine, from + G_SIZE, base_ill)>0)
        reset_cine(p);

    if(p->debug)
        {
        if(ivtc == cine)
            cr=cl='+';
        else if(ivtc == cine + 1)
            cr=cl='-';
        else
            cr=cl='~';
        dbg_group(p, from, from + G_SIZE, base_ill, cine, cl, cr,
            cine + 1, '(', ')');
        }

    /* Output IVTC group */
    return flush_range(m, from, from + G_SIZE, cine + 1);
    }

/* Calculate basic interlacing level of regular frames */
static int group_base(m_priv_t *p, int from)
    {
    int i, j;
    int sort_ill[G_SIZE];
    int base_ill;

    sort_ill[0]=p->frames[from].ill;
    for(i=1; i < G_SIZE; ++i)
        {
        if(p->frames[from+i].ill > sort_ill[i-1])
            sort_ill[i]=p->frames[from+i].ill;
        else
            {
            sort_ill[i]=sort_ill[i-1];
            for(j=i - 1; j > 0; --j)
                {
                if(p->frames[from+i].ill > sort_ill[j-1])
                    break;
                sort_ill[j]=sort_ill[j-1];
                }
            sort_ill[j]=p->frames[from+i].ill;
            }
        }

    base_ill=0;
    for(i=0; i < G_SIZE / 2; ++i)
        base_ill+=sort_ill[i];
    if(sort_ill[i] - sort_ill[i-1] < sort_ill[i+1] - sort_ill[i])
        base_ill=(base_ill + sort_ill[i]) / (G_SIZE / 2 + 1);
    else
        base_ill=base_ill / (G_SIZE / 2);

    return base_ill;
    }

/* Restore telecined frames */
static int group_ivtc(m_priv_t *p, int cine, int base_ill)
    {
    int merge_ill;
    int ret;

    /* Calculate interlacing of combined frame */
    merge_ill=count_il(&p->frames[cine].img, &p->frames[cine+1].img, p->noise);

    /* Select IVTC method */
    if(p->frames[cine].ill * p->cine_huge < merge_ill * CINE_DIV)
        ret=cine; /* Drop second frame */
    else if(p->frames[cine+1].ill * p->cine_huge < merge_ill * CINE_DIV)
        {
        /* Drop first frame */
        vi3x_copy_image(&p->frames[cine].img, &p->frames[cine+1].img);
        ret=cine + 1;
        }
    else
        {
        /* Check interlacing level of IVTC frame */
        if(0 == base_ill || merge_ill * CINE_DIV <= base_ill * p->cine_huge)
            /* Combine frames */
            merge_fields(&p->frames[cine].img, &p->frames[cine+1].img, 0);
        else
            {
            if(p->debug)
                dbg_msg(p, "BAD IVTC: ill[%d]=%d(%d)\n",
                    cine, merge_ill, merge_ill * CINE_DIV / base_ill);
            /* Replace IVTC frame */
            if(cine + 2 < p->count && p->frames[cine+2].ill < merge_ill)
                {
                vi3x_copy_image(&p->frames[cine].img, &p->frames[cine+2].img);
                if(p->debug)
                    dbg_msg(p, "REPLACED BY: ill[%d]=%d(%d)\n",
                        cine + 2, p->frames[cine+2].ill,
                        p->frames[cine+2].ill * CINE_DIV / base_ill);
                }
            else if(cine > 0 && p->frames[cine-1].ill < merge_ill)
                {
                vi3x_copy_image(&p->frames[cine].img, &p->frames[cine-1].img);
                if(p->debug)
                    dbg_msg(p, "REPLACED BY: ill[%d]=%d(%d)\n",
                        cine - 1, p->frames[cine-1].ill,
                        p->frames[cine-1].ill * CINE_DIV / base_ill);
                }
            }
        ret=cine + cine + 1;
        }

    return ret;
    }

/* Fix frames before cine */
static void group_fix_head(m_priv_t *p, int from, int cine, int base_ill)
    {
    int i;

    /* Replace forward */
    for(i=cine - 1; i > from; --i)
        cond_replace_frame(p, i, i - 1, base_ill);

    /* Replace backward */
    for(i=from; i < cine; ++i)
        cond_replace_frame(p, i, i + 1, base_ill);
    }

/* Fix frames after cine */
static int group_fix_tail(m_priv_t *p, int cine, int to, int base_ill)
    {
    int i;
    int ret=0;

    /* Replace backward */
    for(i=cine + 2; i < to - 1; ++i)
        ret+=cond_replace_frame(p, i, i + 1, base_ill);

    /* Replace forward */
    for(i=to - 1; i>cine + 2; --i)
        ret+=cond_replace_frame(p, i, i - 1, base_ill);
    ret+=cond_replace_frame(p, i, cine, base_ill);

    return ret;
    }

/* Replace interlaced frames */
static int cond_replace_frame(m_priv_t *p, int idst, int isrc, int base_ill)
    {
    /* Compare interlacing level */
    if(base_ill > 0
        && p->frames[idst].ill * CINE_DIV > base_ill * p->cine_lo
        && p->frames[idst].ill > p->frames[isrc].ill)
        {
        /* Replace interlaced frame */
        if(p->debug)
            dbg_msg(p, "REPLACE: ill[%d]=%d(%d) BY ill[%d]=%d(%d)\n",
                idst, p->frames[idst].ill,
                p->frames[idst].ill * CINE_DIV / base_ill,
                isrc, p->frames[isrc].ill,
                p->frames[isrc].ill * CINE_DIV / base_ill
                );
        vi3x_copy_image(&p->frames[idst].img, &p->frames[isrc].img);
        p->frames[idst].ill=p->frames[isrc].ill;
        return 1;
        }

    return 0;
    }

/* Search frame to skip */
static int group_pre_skip(m_priv_t *p, int from)
    {
    int i;
    int i_skip=-1;

    if(0==p->pre_cine)
        return i_skip;

    /* If need to skip extra frame */
    if(p->delta < (G_SIZE - 1) / 2)
        {
        /* Search interlacing maximum */
        i_skip=from - p->pre_cine;;
        for(i=i_skip + 1; i < from; ++i)
            {
            if(p->frames[i_skip].ill < p->frames[i].ill)
                i_skip=i;
            }
        p->delta+=G_SIZE - p->pre_cine;
        }
    else
        /* Output all frames */
        p->delta-=p->pre_cine;

    return i_skip;
    }

/* Output pre_cine frames */
static int group_pre_cine(vi3x_module_t *m, int from, int i_skip)
    {
    m_priv_t *p=m->priv;

    if(p->debug)
        {
        dbg_msg(p, "PRE CINE: pre_cine=%d i_skip=%d delta=%d\n",
            p->pre_cine, i_skip, p->delta);
        dbg_group(p, from - p->pre_cine, from,
            i_skip < 0 ? p->frames[from-p->pre_cine].ill : p->frames[i_skip].ill,
            i_skip, '-', '-', -1, 0, 0);
        }

    /* Output frames */
    return flush_range(m, from - p->pre_cine, from, i_skip);
    }

/* Output group of frames */
static int flush_range(vi3x_module_t *m, int from, int to, int skip)
    {
    while(from < to)
        {
        if(from != skip)
            {
            if(!m->out_image(m, &m->priv->frames[from].img))
                return 0;
            ++m->priv->frame_o;
            }
        ++from;
        }

    return 1;
    }

/* Flush last group */
static int flush_tail(vi3x_module_t *m)
    {
    m_priv_t *p=m->priv;
    int cine, drop;
    int ivtc;
    char cl, cr;

    if(m->priv->debug)
        dbg_msg(m->priv, "TAIL: pre_cine=%d cine=%d\n", p->pre_cine, p->cine);

    cine=p->pre_cine + p->cine;
    drop=cine + 1;

    /* Do IVTC */
    if(cine >= p->pre_cine && drop < p->count)
        ivtc=group_ivtc(p, cine, p->prev_ill);
    else
        {
        ivtc=-1;
        drop=-1;
        }

    /* Output pre_cine frames */
    if(p->pre_cine > 0)
        {
        if(!group_pre_cine(m, p->pre_cine, group_pre_skip(p, p->pre_cine)))
            return 0;
        }

    if(p->debug)
        {
        if(ivtc == cine)
            cr=cl='+';
        else if(ivtc == cine + 1)
            cr=cl='-';
        else if(ivtc == cine + cine + 1)
            cr=cl='~';
        else
            cr=cl=' ';
        dbg_group(p, p->pre_cine, p->count, p->frames[p->pre_cine].ill,
            cine, cl, cr, drop, '(', ')');
        }

    /* Output IVTC group */
    return flush_range(m, p->pre_cine, p->count, drop);
    }

typedef struct il_count_s
    {
    int d1, d2, d3;
    int csign_t, csign_b;
    } il_count_t;

static inline void add_il(il_count_t *ilc, int p0, int p1, int p2, int p3)
    {
    int v1, v2;

    v1=p0 - p1;
    v2=p2 - p1;
    ilc->csign_t+=v1 * v2;
    ilc->d1+=v1 * v1;
    ilc->d2+=v2 * v2;
    v1=-v2;
    v2=p3 - p2;
    ilc->csign_b+=v1 * v2;
    ilc->d3+=v2 * v2;
    }

static int count_il(
    vi3x_image_t const *img_bottom, vi3x_image_t const *img_top, int dmin)
    {
    int sum=0;
    int w=img_top->w;
    int count=img_top->count;
    vi3x_sample_t const *yy_t;
    vi3x_sample_t const *yy_b;
    vi3x_sample_t const *end;

    dmin*=NOISE_SCALE;

    yy_t=img_top->yy;
    yy_b=img_bottom->yy + w;
    end=yy_t + count;
    while(yy_t < end)
        {
        vi3x_sample_t const *rend=yy_t + w;
        while(yy_t < rend)
            {
            il_count_t ilc={0, 0, 0, 0, 0};
            add_il(&ilc, yy_t[0], yy_b[0], yy_t[w*2], yy_b[w*2]); /* yy */
            add_il(&ilc, yy_t[count], yy_b[count],
                yy_t[count+w*2], yy_b[count+w*2]); /* cb */
            add_il(&ilc, yy_t[count*2], yy_b[count*2],
                yy_t[count*2+w*2], yy_b[count*2+w*2]); /* cr */
            if(ilc.csign_t > CSIGN_THRESHOLD && ilc.csign_b > CSIGN_THRESHOLD
                && ilc.d1 > dmin && ilc.d2 > dmin && ilc.d3 > dmin)
                ++sum;
            ++yy_b;
            ++yy_t;
            }
        yy_b+=w;
        yy_t+=w;
        }

    return sum;
    }

static void merge_fields(vi3x_image_t *dst, vi3x_image_t const *src, int bottom)
    {
    vi3x_sample_t *to, *end;
    vi3x_sample_t const *from;

    from=src->yy + src->w * bottom;
    to=dst->yy + dst->w * bottom;
    end=dst->yy + dst->count;
    while(to < end)
        {
        memcpy(to, from, dst->w * sizeof(*to));
        to+=dst->w * 2;
        from+=src->w * 2;
        }

    from=src->cb + src->w * bottom;
    to=dst->cb + dst->w * bottom;
    end=dst->cb + dst->count;
    while(to < end)
        {
        memcpy(to, from, dst->w * sizeof(*to));
        to+=dst->w * 2;
        from+=src->w * 2;
        }

    from=src->cr + src->w * bottom;
    to=dst->cr + dst->w * bottom;
    end=dst->cr + dst->count;
    while(to < end)
        {
        memcpy(to, from, dst->w * sizeof(*to));
        to+=dst->w * 2;
        from+=src->w * 2;
        }
    }

static void dbg_msg(m_priv_t const *p, char const *fmt, ...)
    {
    va_list ap;

    vi3x_info("[%s] %06d(%03d): ", Tag, p->frame_o, p->count);
    va_start(ap, fmt);
    vi3x_info_va(fmt, ap);
    va_end(ap);
    }

static void dbg_group(
    m_priv_t const *p,
    int from, int to,
    int scale,
    int i1, char c1_l, char c1_r,
    int i2, char c2_l, char c2_r
    )
    {
    int i;
    int v;
    char cl, cr;

    vi3x_info("[%s] %06d(%03d): [%03d]:", Tag, p->frame_o, p->count, from);
    for(i=from; i < to; ++i)
        {
        v=p->frames[i].ill * CINE_DIV;
        if(0 != scale)
            v=v / scale;
        if(i1 == i)
            {
            cl=c1_l;
            cr=c1_r;
            }
        else if(i2 == i)
            {
            cl=c2_l;
            cr=c2_r;
            }
        else
            cl=cr=' ';
        vi3x_info("\t%c%d%c", cl, v, cr);
        }
    vi3x_info("\t/ %d\n", scale);
    }

