/*
This file is part of video3x.

video3x is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

video3x is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with video3x.  If not, see <http://www.gnu.org/licenses/>.

Copyright 2011 Vadim Druzhin <cdslow@mail.ru>
*/

#include <stdint.h>
#include <stddef.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <limits.h>
#include "vi3x_modules.h"

static char const *Tag="toonclean";
static void (*err)(char const *fmt, ...);
static void (*warn)(char const *fmt, ...);
static void (*info)(char const *fmt, ...);

enum
    {
    Q_YY=0,
    Q_CB,
    Q_CR,
    Q_PLANES
    };

typedef struct levels_s
    {
    int yy;
    int cb;
    int cr;
    } levels_t;

typedef struct qframe_s
    {
    int w;
    int h;
    int qsize;
    vi3x_sample_t *samples;
    levels_t *levels;
    int *attrs;
    } qframe_t;

typedef struct m_priv_s
    {
    int denoise;
    int noise_y;
    int noise_c;
    int depth;
    int depth_first;
    int demo;
    int debug;
    vi3x_image_t *image;
    qframe_t qframe;
    int qlen;
    int qfill;
    int next_in;
    int next_out;
    vi3x_sample_t *non_black;
    vi3x_sample_t *row;
    } m_priv_t;

static void m_help(vi3x_help_t *h);
static void help_denoise(vi3x_arg_t const *arg, vi3x_help_t const *h);
static void help_level(vi3x_arg_t const *arg, vi3x_help_t const *h);
static void help_level_c(vi3x_arg_t const *arg, vi3x_help_t const *h);
static void help_depth(vi3x_arg_t const *arg, vi3x_help_t const *h);
static void help_depth_first(vi3x_arg_t const *arg, vi3x_help_t const *h);
static void help_demo(vi3x_arg_t const *arg, vi3x_help_t const *h);
static void help_debug(vi3x_arg_t const *arg, vi3x_help_t const *h);

static const vi3x_arg_t ArgMap[] = {
    { "denoise",     2, offsetof(m_priv_t, denoise), help_denoise },
    { "level",      -1, offsetof(m_priv_t, noise_y), help_level },
    { "level_c",    -1, offsetof(m_priv_t, noise_c), help_level_c },
    { "depth",       1, offsetof(m_priv_t, depth), help_depth },
    { "depth_first", 0, offsetof(m_priv_t, depth_first), help_depth_first },
    { "demo",        0, offsetof(m_priv_t, demo), help_demo },
    { "debug",       0, offsetof(m_priv_t, debug), help_debug },
    { NULL, 0, 0, NULL }
};

#define DENOISE_MAX 32
#define DENOISE_SCALE 128
/* LN(denoise) * 32.37 + 16 */
static const int DenoiseMap[DENOISE_MAX]=
    {
    16, 38, 52, 61, 68, 74, 79, 83, 87, 91, 94, 96, 99, 101, 104, 106, 108,
    110, 111, 113, 115, 116, 117, 119, 120, 121, 123, 124, 125, 126, 127, 128
    };

/* Noise can't be greater than brightess/10 */
#define BRIGHT_TO_NOISE(x) ((x) / 10)

/* 20% of image considered black */
#define BLACK_COUNT(x) ((x) / 5)
/* Maximum YCBCR distance of black pixels */
#define BLACK_MAX (VI3X_MAX / 2)

/* Take into account 95% of gradients in noise estimation */
#define NOISE_COUNT(x) ((x) * 95 / 100)

typedef struct dist_stat
    {
    int count;
    int freq[VI3X_MAX+1]; /* Minimum value assumed to be 0 */
    } dist_stat_t;

static int filter_init(vi3x_module_t *flt);
static int filter_fini(vi3x_module_t *flt);
static int filter_flush(vi3x_module_t *flt, vi3x_image_t *image);
static int filter_process(vi3x_module_t *flt, vi3x_image_t *image);

VI3X_MAIN()
    {
    vi3x_reg_filter(Tag, filter_init, m_help);
    }

static void reset_fd(m_priv_t *fd);
static int do_frame(m_priv_t *fd);
static int do_last(m_priv_t *fd);
static void frame_preprocess(m_priv_t *fd, int f);
static void frame_preprocess_spatial(m_priv_t *fd, int f);
static void frame_preprocess_temp(m_priv_t *fd, int f);
static void frame_duplicate(m_priv_t *fd, int f);
static void frame_postprocess(m_priv_t *fd, int f);
static void frame_postprocess_spatial(m_priv_t *fd, int f);
static void frame_postprocess_temp(m_priv_t *fd, int f);
static void copy_image(qframe_t *dst, int f, vi3x_image_t const *src);
static void copy_half_image(qframe_t *dst, int f, vi3x_image_t const *src);
static void copy_half_frame(vi3x_image_t *dst, qframe_t const *src, int f);
static void dup_frame(qframe_t *qframe, int to, int from);
static void detect_levels(levels_t *levels, vi3x_image_t const *img,
    vi3x_sample_t *non_black, int debug);
static void black_mask(vi3x_sample_t *non_black, vi3x_image_t const *img);
static int selective_blur(
    vi3x_sample_t *px, int w, int h, vi3x_sample_t *buf, int level);
static void black_distance(
    dist_stat_t *dist, vi3x_sample_t *non_black, vi3x_image_t const *img);
static void distance2mask(vi3x_sample_t *non_black, int w, int h, int level);
static void grow_black(vi3x_sample_t *px, int w, int h);
static int noise_level(
    vi3x_sample_t const *grad, vi3x_sample_t const *non_black, int w, int h);
static void noise_stat(dist_stat_t *stat, vi3x_sample_t const *px,
    vi3x_sample_t const *non_black, int w, int h);
static int stat_select(dist_stat_t const *stat, int amount, int max);
static int brightness(
    vi3x_sample_t const *yy, vi3x_sample_t const *non_black, int w, int h);
static void filter_single(vi3x_image_t *img, levels_t const *levels,
    int denoise, vi3x_sample_t *row, int debug);
static void filter_plane(vi3x_sample_t *plane, int w, int h,
    int level, int denoise, vi3x_sample_t *row, int debug);
static void filter_depth(vi3x_image_t *img, qframe_t const *qframe, int f);
static inline int abs_diff(int a, int b);

/* Module help */
static void m_help(vi3x_help_t *h)
    {
    h->help_module(Tag, VI3X_VER_STR, "Vadim Druzhin",
        "Cartoon-oriented denoise filter",
        "This filter performs a spatial and temporal denoise\n"
        "while preserving sharp edges, also supports an automatic\n"
        "noise level detection.",
        ArgMap);
    }

static void help_denoise(vi3x_arg_t const *arg, vi3x_help_t const *h)
    {
    h->help_param_range(arg->name,
        "Strength of spatial denoise",
        "The 0 value disables spatial denoise, big values will\n"
        "make the processing extremely slow without a visible effect.\n"
        "Recommended values are in the range of 1-4.",
        arg->def, 0, DENOISE_MAX
        );
    }

static void help_level(vi3x_arg_t const *arg, vi3x_help_t const *h)
    {
    h->help_param_range(arg->name,
        "Luma noise level (the pixel value difference)",
        "The -1 value means auto-detect, 0 disables denoise in the luma plane.\n"
        "The typical values are in the range of 4-12.",
        arg->def, -1, VI3X_MAX
        );
    }

static void help_level_c(vi3x_arg_t const *arg, vi3x_help_t const *h)
    {
    h->help_param_range(arg->name,
        "Chroma noise level (the pixel value difference)",
        "The -1 value means auto-detect, 0 disables denoise in the chroma planes.\n"
        "The typical values are in the range of 2-6.",
        arg->def, -1, VI3X_MAX
        );
    }

static void help_depth(vi3x_arg_t const *arg, vi3x_help_t const *h)
    {
    h->help_param_range(arg->name,
        "Depth of the temporal denoise buffer",
        "The buffer combines past and future frames, so with depth=2\n"
        "5 frames will be buffered -- 2 past, 2 future and 1 current.\n"
        "The 0 value disables the temporal denoise.\n"
        "Recommended values are in the range of 1-3.",
        arg->def, 0, INT_MAX
        );
    }

static void help_depth_first(vi3x_arg_t const *arg, vi3x_help_t const *h)
    {
    h->help_param_bool(arg->name,
        "Perform the temporal denoise before the spatial one",
        "By default, the spatial denoise is performed first.",
        arg->def, "", ""
        );
    }

static void help_demo(vi3x_arg_t const *arg, vi3x_help_t const *h)
    {
    h->help_param_bool(arg->name,
        "Process only the right half of a frame", "", arg->def, "", "");
    }

static void help_debug(vi3x_arg_t const *arg, vi3x_help_t const *h)
    {
    h->help_param_bool(arg->name,
        "Enable debug information", "", arg->def, "", "");
    }

static int filter_init(vi3x_module_t *flt)
    {
    m_priv_t params;
    m_priv_t *me;
    uint8_t *buf;
    int sz_row;
    int sz_non_black;
    int sz_qframe;
    int sz_qlevels;
    int sz_attrs;

    err=flt->vi3x.err;
    warn=flt->vi3x.warn;
    info=flt->vi3x.info;

    /* Verify frame size */
    if(flt->w < 3 || flt->h < 3)
        {
        err("[%s] Error: Frame size too small: %d x %d < 3 x 3 !\n",
            Tag, flt->w, flt->h);
        return 0;
        }

    /* Initialize parameters */
    if(!flt->vi3x.parse_args(&params, flt->args, ArgMap, Tag))
        return 0;
    if(params.depth < 0)
        {
        err("[%s] Error: 'depth' can't be negative (depth=%d)!\n",
            Tag, params.depth);
        return 0;
        }
    if(params.denoise < 0 || params.denoise > DENOISE_MAX)
        {
        err("[%s] Error: 'denoise=%d' out of range (%d ... %d)!\n",
            Tag, params.denoise, 0, DENOISE_MAX);
        return 0;
        }
    if(params.denoise>0)
        params.denoise=DenoiseMap[params.denoise-1];
    params.qframe.w=flt->w;
    params.qframe.h=flt->h;
    params.qframe.qsize=params.depth * 2 + 1;

    /* Allocate buffer */
    sz_row=flt->w * sizeof(*me->row);
    sz_non_black=flt->w * flt->h * sizeof(*me->non_black);
    sz_qframe=params.qframe.w * params.qframe.h
        * Q_PLANES * params.qframe.qsize * sizeof(*params.qframe.samples);
    sz_qlevels=params.qframe.qsize * sizeof(*params.qframe.levels);
    sz_attrs=params.qframe.qsize * sizeof(*params.qframe.attrs);
    buf=flt->vi3x.alloc(sizeof(*me) + sz_row + sz_non_black + sz_qframe
        + sz_qlevels + sz_attrs, Tag);

    /* Store parameters */
    me=(m_priv_t *)buf;
    buf+=sizeof(*me);
    *me=params;

    /* Initialize state */
    reset_fd(me);

    /* Initialize buffer pointers */
    me->qframe.samples=(vi3x_sample_t *)buf;
    buf+=sz_qframe;
    me->qframe.levels=(levels_t *)buf;
    buf+=sz_qlevels;
    me->qframe.attrs=(int *)buf;
    buf+=sz_attrs;
    me->row=(vi3x_sample_t *)buf;
    buf+=sz_row;
    me->non_black=(vi3x_sample_t *)buf;
    buf+=sz_non_black;

    /* Initialize methods */
    flt->process_image=filter_process;
    flt->flush=filter_flush;
    flt->fini=filter_fini;

    /* Store private data */
    flt->priv=me;

    return 1;
    }

static int filter_fini(vi3x_module_t *flt)
    {
    if(NULL!=flt->priv)
        free(flt->priv);

    return 1;
    }

static int filter_process(vi3x_module_t *flt, vi3x_image_t *image)
    {
    flt->priv->image=image;
    if(do_frame(flt->priv))
        return flt->out_image(flt, flt->priv->image);

    return 1;
    }

static int filter_flush(vi3x_module_t *flt, vi3x_image_t *image)
    {
    int ret=1;

    flt->priv->image=image;
    while(ret && do_last(flt->priv))
        ret=flt->out_image(flt, flt->priv->image);

    reset_fd(flt->priv);

    return ret;
    }


static void reset_fd(m_priv_t *fd)
    {
    fd->qlen=0;
    fd->qfill=0;
    fd->next_in=0;
    fd->next_out=0;
    }

static int do_frame(m_priv_t *fd)
    {
    /* Process image */
    frame_preprocess(fd, fd->next_in);
    fd->next_in=(fd->next_in + 1) % fd->qframe.qsize;
    ++fd->qlen;
    ++fd->qfill;

    if(fd->qfill == fd->qframe.qsize)
        {
        /* Fill frame */
        frame_postprocess(fd, fd->next_out);
        fd->next_out=(fd->next_out + 1) % fd->qframe.qsize;
        --fd->qlen;
        --fd->qfill;
        return 1;
        }

    /* Repeat first frame */
    if(1 == fd->qlen)
        {
        while(fd->qfill <= fd->qframe.qsize / 2)
            {
            frame_duplicate(fd, fd->next_in);
            ++fd->next_in;
            ++fd->next_out;
            ++fd->qfill;
            }
        }

    return 0;
    }

static int do_last(m_priv_t *fd)
    {
    if(0 == fd->qlen)
        return 0;

    /* Repeat last frame */
    while(fd->qfill < fd->qframe.qsize)
        {
        frame_duplicate(fd, fd->next_in);
        fd->next_in=(fd->next_in + 1) % fd->qframe.qsize;
        ++fd->qfill;
        }
    /* Fill frame */
    frame_postprocess(fd, fd->next_out);
    fd->next_out=(fd->next_out + 1) % fd->qframe.qsize;
    --fd->qlen;

    return 1;
    }

static void frame_preprocess(m_priv_t *fd, int f)
    {
    if(fd->depth_first)
        /* Temporal first */
        frame_preprocess_temp(fd, f);
    else
        /* Spatial first */
        frame_preprocess_spatial(fd, f);
    }

static void frame_postprocess(m_priv_t *fd, int f)
    {
    if(fd->depth_first)
        /* Temporal first */
        frame_postprocess_temp(fd, f);
    else
        /* Spatial first */
        frame_postprocess_spatial(fd, f);
    }

/* Denoise spatial first, pass 1 */
static void frame_preprocess_spatial(m_priv_t *fd, int f)
    {
    /* Initialize levels */
    fd->qframe.levels[f].yy=fd->noise_y;
    fd->qframe.levels[f].cb=fd->noise_c;
    fd->qframe.levels[f].cr=fd->noise_c;

    /* Adaptive levels */
    if(fd->noise_y < 0 || fd->noise_c < 0)
        detect_levels(fd->qframe.levels + f, fd->image, fd->non_black, fd->debug);

    /* Save half of original image */
    if(fd->demo && fd->denoise)
        copy_half_image(&fd->qframe, f, fd->image);

    /* Spatial denoise */
    if(fd->denoise)
        filter_single(fd->image, fd->qframe.levels + f, fd->denoise,
            fd->row, fd->debug);

    /* Restore half of original image */
    if(fd->demo && fd->denoise)
        copy_half_frame(fd->image, &fd->qframe, f);

    /* Copy source image */
    if(fd->qframe.qsize > 1)
        copy_image(&fd->qframe, f, fd->image);
    }

/* Denoise spatial first, pass 2 */
static void frame_postprocess_spatial(m_priv_t *fd, int f)
    {
    if(fd->qframe.qsize > 1)
        {
        /* Image temporal filtering */
        filter_depth(fd->image, &fd->qframe, f);

        /* Restore half of original image */
        if(fd->demo)
            copy_half_frame(fd->image, &fd->qframe, f);
        }
    }

/* Denoise temporal first, pass 1 */
static void frame_preprocess_temp(m_priv_t *fd, int f)
    {
    /* Initialize levels */
    fd->qframe.levels[f].yy=fd->noise_y;
    fd->qframe.levels[f].cb=fd->noise_c;
    fd->qframe.levels[f].cr=fd->noise_c;

    /* Adaptive levels */
    if(fd->noise_y < 0 || fd->noise_c < 0)
        detect_levels(fd->qframe.levels + f, fd->image, fd->non_black, fd->debug);

    /* Copy source image */
    if(fd->qframe.qsize > 1)
        copy_image(&fd->qframe, f, fd->image);
    }

/* Denoise temporal first, pass 2 */
static void frame_postprocess_temp(m_priv_t *fd, int f)
    {
    /* Image temporal filtering */
    if(fd->qframe.qsize > 1)
        filter_depth(fd->image, &fd->qframe, f);

    /* Spatial denoise */
    if(fd->denoise)
        {
        /* Save half of original image */
        if(fd->demo && 1 == fd->qframe.qsize)
            copy_half_image(&fd->qframe, f, fd->image);

        filter_single(fd->image, fd->qframe.levels + f, fd->denoise,
            fd->row, fd->debug);
        }

    /* Restore half of original image */
    if(fd->demo)
        copy_half_frame(fd->image, &fd->qframe, f);
    }

static void frame_duplicate(m_priv_t *fd, int f)
    {
    dup_frame(&fd->qframe, f, (f - 1 + fd->qframe.qsize) % fd->qframe.qsize);
    }

static void copy_image(qframe_t *dst, int f, vi3x_image_t const *src)
    {
    vi3x_sample_t *d;
    int xy;

    d=dst->samples + f * Q_PLANES;
    for(xy=0; xy < src->count; ++xy)
        {
        d[Q_YY]=src->yy[xy];
        d[Q_CB]=src->cb[xy];
        d[Q_CR]=src->cr[xy];
        d+=dst->qsize * Q_PLANES;
        }
    dst->attrs[f]=src->il;
    }

static void copy_half_image(qframe_t *dst, int f, vi3x_image_t const *src)
    {
    vi3x_sample_t *d;
    int y, xy;

    for(y=0; y < src->h; ++y)
        {
        d=dst->samples + y * dst->w * dst->qsize * Q_PLANES + f * Q_PLANES;
        for(xy=y * src->w; xy < y * src->w + src->w / 2; ++xy)
            {
            d[Q_YY]=src->yy[xy];
            d[Q_CB]=src->cb[xy];
            d[Q_CR]=src->cr[xy];
            d+=dst->qsize * Q_PLANES;
            }
        }
    }

static void copy_half_frame(vi3x_image_t *dst, qframe_t const *src, int f)
    {
    vi3x_sample_t const *s;
    int xy, y;

    for(y=0; y < dst->h; ++y)
        {
        s=src->samples + y * src->w * src->qsize * Q_PLANES + f * Q_PLANES;
        for(xy=y * dst->w; xy < y * dst->w + dst->w / 2; ++xy)
            {
            dst->yy[xy]=s[Q_YY];
            dst->cb[xy]=s[Q_CB];
            dst->cr[xy]=s[Q_CR];
            s+=src->qsize * Q_PLANES;
            }
        }
    }

static void dup_frame(qframe_t *qframe, int to, int from)
    {
    vi3x_sample_t *s, *end;
    int delta;

    s=qframe->samples + from * Q_PLANES;
    end=qframe->samples + qframe->w * qframe->h * qframe->qsize * Q_PLANES;
    delta=(to - from) * Q_PLANES;

    while(s < end)
        {
        s[delta]=*s;
        ++s;
        s[delta]=*s;
        ++s;
        s[delta]=*s;
        ++s;
        s+=(qframe->qsize - 1) * Q_PLANES;
        }

    qframe->levels[to]=qframe->levels[from];
    qframe->attrs[to]=qframe->attrs[from];
    }

static void detect_levels(levels_t *levels, vi3x_image_t const *img,
    vi3x_sample_t *non_black, int debug)
    {
    int bright, noise;

    /* Mark black pixels */
    black_mask(non_black, img);
    /* Expand black mask by one pixel in each direction */
    grow_black(non_black, img->w, img->h);

    /* Luma plane */
    if(levels->yy < 0)
        {
        /* Adaptive denoise level */
        noise=noise_level(img->yy, non_black, img->w, img->h);
        /* Calculate brightness of non-black areas */
        bright=brightness(img->yy, non_black, img->w, img->h);
        /* Heuristic: maximum luma noise level depends on brightness */
        bright=BRIGHT_TO_NOISE(bright);
        if(noise > bright)
            levels->yy=bright;
        else
            levels->yy=noise;
        }

    /* Blue chroma plane */
    if(levels->cb < 0)
        {
        /* Adaptive denoise level */
        noise=noise_level(img->cb, non_black, img->w, img->h);
        levels->cb=noise;
        }

    /* Red chroma plane */
    if(levels->cr<0)
        {
        /* Adaptive denoise level */
        noise=noise_level(img->cr, non_black, img->w, img->h);
        levels->cr=noise;
        }

    if(debug)
        info("[%s] detect_levels: LUMA=%d BLUE=%d RED=%d\n", Tag,
            levels->yy, levels->cb, levels->cr);
    }

static void black_mask(vi3x_sample_t *non_black, vi3x_image_t const *img)
    {
    dist_stat_t stat;
    int level;

    black_distance(&stat, non_black, img);
    /* Heuristic: some part of pixels must be black,
       absolute maximum of black level is BLACK_MAX */
    level=stat_select(&stat, BLACK_COUNT(stat.count), BLACK_MAX);
    distance2mask(non_black, img->w, img->h, level);
    }

static void filter_single(vi3x_image_t *img, levels_t const *levels,
    int denoise, vi3x_sample_t *row, int debug)
    {
    if(levels->yy > 0)
        {
        if(debug)
            info("[%s] LUMA: ", Tag);
        filter_plane(img->yy, img->w, img->h, levels->yy, denoise, row, debug);
        }

    if(levels->cb>0)
        {
        if(debug)
            info("[%s] BLUE: ", Tag);
        filter_plane(img->cb, img->w, img->h, levels->cb, denoise, row, debug);
        }

    if(levels->cr>0)
        {
        if(debug)
            info("[%s] RED: ", Tag);
        filter_plane(img->cr, img->w, img->h, levels->cr, denoise, row, debug);
        }
    }

static void filter_plane(vi3x_sample_t *plane, int w, int h, int level,
    int denoise, vi3x_sample_t *row, int debug)
    {
    int prev;
    int sqd;

    if(debug)
        info("filter_plane(denoise=%.2f, level=%d): ",
            (double)denoise / (double)DENOISE_SCALE, level);

    sqd=selective_blur(plane, w, h, row, level);
    if(debug)
        info("sqd=%.4f", (double)sqd / (double)(w * h));

    do
        {
        prev=sqd;
        sqd=selective_blur(plane, w, h, row, level);
        if(debug)
            info(", %.4f", (double)sqd / (double)(w * h));
        }
    while(sqd>0 && sqd * DENOISE_SCALE < prev * denoise);

    if(debug)
        info("\n");
    }

static inline vi3x_sample_t yuv_distance(int yy, int cb, int cr)
    {
    cb=(cb - VI3X_GREY) * 2;
    cb*=cb;
    cr=(cr - VI3X_GREY) * 2;
    cr*=cr;
    return (vi3x_sample_t)sqrt((float)((yy * yy + cr + cb) / 3));
    }

static void black_distance(
    dist_stat_t *dist, vi3x_sample_t *non_black, vi3x_image_t const *img)
    {
    int xy;

    memset(dist->freq, 0, sizeof(dist->freq));
    dist->count=img->count;
    for(xy=0; xy < img->count; ++xy)
        {
        non_black[xy]=yuv_distance(img->yy[xy], img->cb[xy], img->cr[xy]);
        ++dist->freq[non_black[xy]];
        }
    }

static void distance2mask(vi3x_sample_t *non_black, int w, int h, int level)
    {
    vi3x_sample_t *end;

    for(end=non_black + w * h; non_black < end; ++non_black)
        {
        if(*non_black <= level)
            *non_black=0;
        else
            *non_black=1;
        }
    }

static int brightness(
    vi3x_sample_t const *yy, vi3x_sample_t const *non_black, int w, int h)
    {
    int sum=0;
    int count=0;
    vi3x_sample_t const *end;

    for(end=yy + w * h; yy < end; ++yy)
        {
        if(*non_black)
            {
            sum+=*yy;
            ++count;
            }
        ++non_black;
        }

    if(0 == count)
        return 0;

    return sum/count;
    }

static int noise_level(
    vi3x_sample_t const *px, vi3x_sample_t const *non_black, int w, int h)
    {
    dist_stat_t stat;

    noise_stat(&stat, px, non_black, w, h);

    /* Heuristic: take into account some part of gradients */
    return stat_select(&stat, NOISE_COUNT(stat.count), VI3X_MAX);
    }

static void noise_stat(dist_stat_t *stat, vi3x_sample_t const *px,
    vi3x_sample_t const *non_black, int w, int h)
    {
    stat->count=0;
    memset(stat->freq, 0, sizeof(stat->freq));

    for(; h > 0; --h)
        {
        vi3x_sample_t const *rend=px + w;
        int px1, px2;
        px2=*px++;
        px1=*px++;
        ++non_black;
        while(px < rend)
            {
            if(*non_black++)
                {
                ++stat->count;
                ++stat->freq[abs_diff(px2, *px)];
                }
            px2=px1;
            px1=*px++;
            }
        ++non_black;
        }
    }

static int stat_select(dist_stat_t const *stat, int amount, int max)
    {
    int level;

    if(0 == stat->count)
        return 0;

    amount-=stat->freq[0];
    for(level=1; level <= max; ++level)
        {
        amount-=stat->freq[level];
        if(amount <= 0)
            break;
        }

    return level - 1;
    }

#define MIN_WEIGHT 5

static inline void cond_sum(int *sum, int *n, int level, int p0, int p)
    {
    if(abs(p0 - p) <= level) {*sum+=p; ++*n;}
    }

static inline void px_avg(vi3x_sample_t *px, int *sqd, int sum, int n, int p0)
    {
    sum=(sum + n / 2) / n;
    *px=(vi3x_sample_t)sum;
    sum-=p0;
    *sqd+=sum * sum;
    }

static int selective_blur(
    vi3x_sample_t *px, int w, int h, vi3x_sample_t *buf, int level)
    {
    vi3x_sample_t *end;
    vi3x_sample_t *rowb, *bend1;
    int rb_0, r1_0, r1_1;
    int sum, n;
    int sqd=0;

    end=px + w * h;
    px+=w;
    bend1=buf + w - 1;

    /*** First row ***/

    n=1;
    r1_0=px[-w]; /* Load first pixel in row */
    sum=r1_0;                              /* Add row1[0] */
    rowb=buf; /* Initialize buffer pointer */
    *rowb=r1_0; /* Move row1 to rowb */
    ++rowb;
    cond_sum(&sum, &n, level, r1_0, *px);  /* Add row2[0] */
    ++px;
    cond_sum(&sum, &n, level, r1_0, *px);  /* Add row2[1] */
    r1_1=px[-w];
    cond_sum(&sum, &n, level, r1_0, r1_1); /* Add row1[1] */
    px_avg(px - w - 1, &sqd, sum, n, r1_0); /* Store selective average */
    while(rowb < bend1)
        {
        n=1;
        sum=r1_1;                                /* Add row1[0] */
        cond_sum(&sum, &n, level, r1_1, r1_0);   /* Add row1[-1] */
        *rowb=r1_1; /* Move row1 to rowb */
        ++rowb;
        cond_sum(&sum, &n, level, r1_1, px[-1]); /* Add row2[-1] */
        cond_sum(&sum, &n, level, r1_1, *px);    /* Add row2[0] */
        ++px;
        cond_sum(&sum, &n, level, r1_1, *px);    /* Add row2[1] */
        r1_0=r1_1;
        r1_1=px[-w];
        cond_sum(&sum, &n, level, r1_0, r1_1);   /* Add row1[1] */
        px_avg(px - w - 1, &sqd, sum, n, r1_0); /* Store selective average */
        }
    n=1;
    sum=r1_1;                                /* Add row1[0] */
    cond_sum(&sum, &n, level, r1_1, r1_0);   /* Add row1[-1] */
    *rowb=r1_1; /* Move row1 to rowb */
    cond_sum(&sum, &n, level, r1_1, px[-1]); /* Add row2[-1] */
    cond_sum(&sum, &n, level, r1_1, *px);    /* Add row2[0] */
    px_avg(px-w, &sqd, sum, n, r1_1); /* Store selective average */
    ++px;

    /*** Other rows ***/

    while(px < end)
        {
        n=1;
        r1_0=px[-w]; /* Load first pixel in row */
        sum=r1_0;                               /* Add row1[0] */
        rowb=buf; /* Initialize buffer pointer */
        rb_0=*rowb; /* Remember old rowb */
        cond_sum(&sum, &n, level, r1_0, rb_0);  /* Add rowb[0] */
        *rowb=r1_0; /* Move row1 to rowb */
        ++rowb;
        cond_sum(&sum, &n, level, r1_0, *rowb); /* Add rowb[1] */
        cond_sum(&sum, &n, level, r1_0, *px);   /* Add row2[0] */
        ++px;
        cond_sum(&sum, &n, level, r1_0, *px);   /* Add row2[1] */
        r1_1=px[-w];
        cond_sum(&sum, &n, level, r1_0, r1_1);  /* Add row1[1] */
        px_avg(px - w - 1, &sqd, sum, n, r1_0); /* Store selective average */
        while(rowb < bend1)
            {
            n=1;
            sum=r1_1;                                /* Add row1[0] */
            cond_sum(&sum, &n, level, r1_1, r1_0);   /* Add row1[-1] */
            cond_sum(&sum, &n, level, r1_1, rb_0);   /* Add rowb[-1] */
            rb_0=*rowb; /* Remember old rowb */
            cond_sum(&sum, &n, level, r1_1, rb_0);   /* Add rowb[0] */
            *rowb=r1_1; /* Move row1 to rowb */
            ++rowb;
            cond_sum(&sum, &n, level, r1_1, *rowb);  /* Add rowb[1] */
            cond_sum(&sum, &n, level, r1_1, px[-1]); /* Add row2[-1] */
            cond_sum(&sum, &n, level, r1_1, *px);    /* Add row2[0] */
            ++px;
            cond_sum(&sum, &n, level, r1_1, *px);    /* Add row2[1] */
            r1_0=r1_1;
            r1_1=px[-w];
            cond_sum(&sum, &n, level, r1_0, r1_1);   /* Add row1[1] */
            /* Store selective average */
            if(n > MIN_WEIGHT)
                px_avg(px - w - 1, &sqd, sum, n, r1_0);
            }
        n=1;
        sum=r1_1;                                /* Add row1[0] */
        cond_sum(&sum, &n, level, r1_1, r1_0);   /* Add row1[-1] */
        cond_sum(&sum, &n, level, r1_1, rb_0);   /* Add rowb[-1] */
        cond_sum(&sum, &n, level, r1_1, *rowb);  /* Add rowb[0] */
        *rowb=r1_1; /* Move row1 to rowb */
        cond_sum(&sum, &n, level, r1_1, px[-1]); /* Add row2[-1] */
        cond_sum(&sum, &n, level, r1_1, *px);    /* Add row2[0] */
        px_avg(px - w, &sqd, sum, n, r1_1); /* Store selective average */
        ++px;
        }

    /*** Last row ***/

    px-=w;
    n=1;
    r1_0=*px; /* Load first pixel in row */
    sum=r1_0;                               /* Add row1[0] */
    rowb=buf; /* Initialize buffer pointer */
    cond_sum(&sum, &n, level, r1_0, *rowb); /* Add rowb[0] */
    ++rowb;
    cond_sum(&sum, &n, level, r1_0, *rowb); /* Add rowb[1] */
    r1_1=px[1];
    cond_sum(&sum, &n, level, r1_0, r1_1);  /* Add row1[1] */
    px_avg(px, &sqd, sum, n, r1_0); /* Store selective average */
    ++px;
    while(rowb < bend1)
        {
        n=1;
        sum=r1_1;                                  /* Add row1[0] */
        cond_sum(&sum, &n, level, r1_1, r1_0);     /* Add row1[-1] */
        cond_sum(&sum, &n, level, r1_1, rowb[-1]); /* Add rowb[-1] */
        cond_sum(&sum, &n, level, r1_1, *rowb);    /* Add rowb[0] */
        ++rowb;
        cond_sum(&sum, &n, level, r1_1, *rowb);    /* Add rowb[1] */
        r1_0=r1_1;
        r1_1=px[1];
        cond_sum(&sum, &n, level, r1_0, r1_1);     /* Add row1[1] */
        px_avg(px, &sqd, sum, n, r1_0); /* Store selective average */
        ++px;
        }
    n=1;
    sum=r1_1;                                  /* Add row1[0] */
    cond_sum(&sum, &n, level, r1_1, r1_0);     /* Add row1[-1] */
    cond_sum(&sum, &n, level, r1_1, rowb[-1]); /* Add rowb[-1] */
    cond_sum(&sum, &n, level, r1_1, *rowb);    /* Add rowb[0] */
    px_avg(px, &sqd, sum, n, r1_1); /* Store selective average */

    return sqd;
    }

static inline int abs_diff(int a, int b)
    {
    return abs(a - b);
    }

static inline void cond_sum3(int *sum, int *n,
    vi3x_sample_t const *s, vi3x_sample_t const *s0, levels_t const *levels)
    {
    if(abs_diff(s[Q_YY], s0[Q_YY]) <= levels->yy
        && abs_diff(s[Q_CB], s0[Q_CB]) <= levels->cb
        && abs_diff(s[Q_CR], s0[Q_CR]) <= levels->cr
        )
        {
        sum[Q_YY]+=s[Q_YY];
        sum[Q_CB]+=s[Q_CB];
        sum[Q_CR]+=s[Q_CR];
        ++*n;
        }
    }

static void filter_depth(vi3x_image_t *img, qframe_t const *qframe, int f)
    {
    levels_t const levels=qframe->levels[f];
    vi3x_sample_t const *src;
    int xy;

    src=qframe->samples;
    for(xy=0; xy < img->count; ++xy)
        {
        vi3x_sample_t const *cmp=src + f * Q_PLANES;
        vi3x_sample_t const *end=src + qframe->qsize * Q_PLANES;
        int n=1;
        int sum[Q_PLANES];
        sum[Q_YY]=cmp[Q_YY];
        sum[Q_CB]=cmp[Q_CB];
        sum[Q_CR]=cmp[Q_CR];
        while(src < cmp)
            {
            cond_sum3(sum, &n, src, cmp, &levels);
            src+=Q_PLANES;
            }
        src+=Q_PLANES;
        while(src<end)
            {
            cond_sum3(sum, &n, src, cmp, &levels);
            src+=Q_PLANES;
            }
        img->yy[xy]=(sum[Q_YY] + n / 2) / n;
        img->cb[xy]=(sum[Q_CB] + n / 2) / n;
        img->cr[xy]=(sum[Q_CR] + n / 2) / n;
        }

    img->il=qframe->attrs[f];
    }

static void grow_black(vi3x_sample_t *px, int w, int h)
    {
    vi3x_sample_t *end=px + w * h;
    int y;

    while(px < end)
        {
        vi3x_sample_t *rend=px + w;
        vi3x_sample_t p0=*px++;
        while(px < rend)
            {
            if(*px != p0)
                {
                p0=*px;
                if(p0)
                    px[0]=0;
                else
                    px[-1]=0;
                }
            ++px;
            }
        }

    for(y=0; y < w; ++y)
        {
        vi3x_sample_t p0;
        px=end - w * h + y;
        p0=*px;
        px+=w;
        while(px < end)
            {
            if(*px != p0)
                {
                p0=*px;
                if(p0)
                    px[0]=0;
                else
                    px[-w]=0;
                }
            px+=w;
            }
        }
    }
