/*
This file is part of video3x.

video3x is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

video3x is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with video3x.  If not, see <http://www.gnu.org/licenses/>.

Copyright 2011 Vadim Druzhin <cdslow@mail.ru>
*/

#include <stdint.h>
#include <string.h>
#include <stddef.h>
#include <limits.h>
#include <stdlib.h>
#include "vi3x_modules.h"

static char const *Tag="toonframe";

#define Q_DEPTH 3
typedef struct m_priv_s
    {
    int mode;
    int noise;
    int motion;
    int force;
    int min_ill;
    int grow_mask;
    int debug;
    int il_warned;
    vi3x_image_t frames[Q_DEPTH];
    int q_len;
    int q_in;
    int q_out;
    int q_fill;
    } m_priv_t;

#define DEINT_SHOW   0
#define DEINT_SMART  1
#define DEINT_SIMPLE 2
#define DEINT_BLEND  3

#define CSIGN_THRESHOLD -4

#define NOISE_SCALE 16

#define ILL_SCALE 1000

#define MARK_YY VI3X_MAX
#define MARK_CB VI3X_GREY
#define MARK_CR VI3X_GREY

#define MAX_DELTA ((VI3X_MAX - VI3X_MIN) / 32)
#define SEARCH_WIDTH 6

static int m_init(vi3x_module_t *m);
static int m_fini(vi3x_module_t *m);
static int m_process(vi3x_module_t *m, vi3x_image_t *image);
static int m_flush(vi3x_module_t *m, vi3x_image_t *image);
static void reset_q(m_priv_t *p);
static int do_frame(vi3x_module_t *m, vi3x_image_t *image);
static void mark_clear(vi3x_image_t *mark);
static void mark_all(vi3x_image_t *mark);
static int mark_il(vi3x_image_t *mark, vi3x_image_t const *img, int noise);
static int unmark_static(
    vi3x_image_t *mark, vi3x_image_t const frames[Q_DEPTH], int motion, int ill);
static void grow_il(vi3x_image_t *mark, int grow);
static void merge_plane(
    vi3x_sample_t *dst, vi3x_sample_t const *src, int w, int h);
static void blend_plane(
    vi3x_sample_t *dst, vi3x_sample_t const *src, int w, int h);
static inline int average(int a, int b);
static void m_help(vi3x_help_t *h);
static void help_mode(vi3x_arg_t const *arg, vi3x_help_t const *h);
static void help_noise(vi3x_arg_t const *arg, vi3x_help_t const *h);
static void help_motion(vi3x_arg_t const *arg, vi3x_help_t const *h);
static void help_force(vi3x_arg_t const *arg, vi3x_help_t const *h);
static void help_debug(vi3x_arg_t const *arg, vi3x_help_t const *h);
static void help_min_ill(vi3x_arg_t const *arg, vi3x_help_t const *h);
static void help_grow_mask(vi3x_arg_t const *arg, vi3x_help_t const *h);
static int smart_avg(
    vi3x_sample_t const *px_u, vi3x_sample_t const *px_l);
static void deint_simple(
    vi3x_sample_t *dst, vi3x_sample_t const *src, int w, int h);
static void deint_smart(
    vi3x_sample_t *dst, vi3x_sample_t const *src, int w, int h);
static inline int abs_diff(int a, int b);

static const vi3x_arg_t ArgMap[] =
    {
    { "mode", DEINT_SMART, offsetof(m_priv_t, mode), help_mode },
    { "noise",      8, offsetof(m_priv_t, noise), help_noise },
    { "motion",    64, offsetof(m_priv_t, motion), help_motion },
    { "force",      0, offsetof(m_priv_t, force), help_force },
    { "min_ill",    0, offsetof(m_priv_t, min_ill), help_min_ill },
    { "grow_mask",  2, offsetof(m_priv_t, grow_mask), help_grow_mask },
    { "debug",      0, offsetof(m_priv_t, debug), help_debug },
    { NULL, 0, 0, NULL }
    };

VI3X_MAIN()
    {
    vi3x_reg_filter(Tag, m_init, m_help);
    }

/* Module help */
static void m_help(vi3x_help_t *h)
    {
    h->help_module(Tag, VI3X_VER_STR, "Vadim Druzhin",
        "Cartoon-oriented area based deinterlace filter",
        "This filter interpolates or blends fields\n"
        "in moving areas of an interlaced video.",
        ArgMap
        );
    }

static void help_mode(vi3x_arg_t const *arg, vi3x_help_t const *h)
    {
    h->help_param_list(arg->name,
        "Deinterlacing mode", "",
        arg->def, 4,
        DEINT_SHOW,   "- just mark interlaced areas;",
        DEINT_SMART,  "- smart linear interpolation;",
        DEINT_SIMPLE, "- simple linear interpolation;",
        DEINT_BLEND,  "- blend fields together."
        );
    }

static void help_noise(vi3x_arg_t const *arg, vi3x_help_t const *h)
    {
    h->help_param_range(arg->name,
        "Noise threshold for the interlacing detector",
        "If you see too much false interlacing, increase this value.",
        arg->def, 0, VI3X_MAX * VI3X_MAX * 3 / NOISE_SCALE
        );
    }

static void help_motion(vi3x_arg_t const *arg, vi3x_help_t const *h)
    {
    h->help_param_range(arg->name,
        "Motion detection sensitivity",
        "With smaller values more areas will be consider as moving\n"
        "and get deinterlaced. The 0 value disables motion detection.",
        arg->def, 0, VI3X_MAX * VI3X_MAX
        );
    }

static void help_force(vi3x_arg_t const *arg, vi3x_help_t const *h)
    {
    h->help_param_bool(arg->name,
        "Disable the interlacing detection",
        "All moving areas will be deinterlaced.",
        arg->def, "", ""
        );
    }

static void help_min_ill(vi3x_arg_t const *arg, vi3x_help_t const *h)
    {
    h->help_param_range(arg->name,
        "Minimum interlacing level to perform deinterlacing",
        "The 0 value means deinterlace all frames.",
        arg->def, 0, ILL_SCALE
        );
    }

static void help_grow_mask(vi3x_arg_t const *arg, vi3x_help_t const *h)
    {
    h->help_param_range(arg->name,
        "Expand detected interlaced area by this amount of pixels", "",
        arg->def, 0, INT_MAX
        );
    }

static void help_debug(vi3x_arg_t const *arg, vi3x_help_t const *h)
    {
    h->help_param_bool(arg->name,
        "Enable debug information", "", arg->def, "", "");
    }

static int m_init(vi3x_module_t *m)
    {
    m_priv_t tmp;
    uint8_t *buf;
    int i;

    /* Check frame parameters */
    if(m->h < 4)
        {
        m->vi3x.err("[%s] Error: Frame height too small: h=%d < 4 !\n", Tag, m->h);
        return 0;
        }
    if(m->h & 1)
        {
        m->vi3x.err("[%s] Error: Uneven frame height: h=%d !\n", Tag, m->h);
        return 0;
        }

    /* Initialize module parameters */
    if(!m->vi3x.parse_args(&tmp, m->args, ArgMap, Tag))
        return 0;

    if(DEINT_SMART == tmp.mode && m->w < SEARCH_WIDTH * 2)
        {
        m->vi3x.err("[%s] Error: Frame width too small: w=%d < %d !\n",
            Tag, m->w, SEARCH_WIDTH * 2);
        return 0;
        }

    /* Calculate image buffer size */
    m->vi3x.init_image(&tmp.frames[0], m->w, m->h, NULL);

    /* Allocate buffer */
    buf=m->vi3x.alloc(sizeof(*m->priv) + tmp.frames[0].img_sz * Q_DEPTH, Tag);

    /* Store parameters */
    m->priv=(m_priv_t *)buf;
    buf+=sizeof(*m->priv);
    *m->priv=tmp;

    /* Initialize buffer pointers */
    for(i=0; i < Q_DEPTH; ++i)
        buf=(uint8_t *)m->vi3x.init_image(
            m->priv->frames + i, m->w, m->h, (vi3x_sample_t *)buf);

    /* Initialize state */
    reset_q(m->priv);
    m->priv->il_warned=0;

    /* Publish module interface */
    m->fini=m_fini;
    m->process_image=m_process;
    m->flush=m_flush;

    return 1;
    }

static void reset_q(m_priv_t *p)
    {
    p->q_in=0;
    p->q_out=0;
    p->q_len=0;
    p->q_fill=0;
    }

static int m_fini(vi3x_module_t *m)
    {
    if(NULL != m->priv)
        m->vi3x.free(m->priv);

    return 1;
    }

static int m_process(vi3x_module_t *m, vi3x_image_t *image)
    {
    m_priv_t *p=m->priv;

    /* Verify frame type */
    if(!image->il && !p->il_warned)
        {
        m->vi3x.warn("[%s] Warning: Input frame marked as progressive, "
            "processing may be inaccurate!\n", Tag);
        p->il_warned=1;
        }

    m->vi3x.copy_image(p->frames + p->q_in, image);
    p->q_in=(p->q_in + 1) % Q_DEPTH;
    ++p->q_len;
    ++p->q_fill;

    if(Q_DEPTH == p->q_fill)
        return do_frame(m, image);

    if(1 == p->q_len)
        {
        /* Duplicate first frame */
        m->vi3x.copy_image(p->frames + p->q_in, image);
        ++p->q_in;
        ++p->q_out;
        ++p->q_fill;
        }

    return 1;
    }

static int m_flush(vi3x_module_t *m, vi3x_image_t *image)
    {
    m_priv_t *p=m->priv;
    int ret=1;

    while(ret && p->q_len > 0)
        {
        while(p->q_fill < Q_DEPTH)
            {
            m->vi3x.copy_image(p->frames + p->q_in,
                p->frames + (p->q_in - 1 + Q_DEPTH) % Q_DEPTH);
            p->q_in=(p->q_in + 1) % Q_DEPTH;
            ++p->q_fill;
            }
        ret=do_frame(m, image);
        }

    reset_q(m->priv);

    return ret;
    }

static int do_frame(vi3x_module_t *m, vi3x_image_t *image)
    {
    m_priv_t *p=m->priv;
    int ill=0;

    if(p->force)
        {
        mark_all(image);
        ill=image->w * image->h;
        }
    else
        {
        mark_clear(image);
        ill=mark_il(image, p->frames + p->q_out, p->noise);
        }

    if(p->motion > 0)
        ill=unmark_static(image, p->frames, p->motion, ill);

    if(p->debug)
        m->vi3x.info("[%s] %06ld: ill=%03d", Tag, m->frame_n - 1,
            ill * ILL_SCALE / (image->w * image->h));

    if(DEINT_SHOW == p->mode)
        {
        if(p->debug)
            m->vi3x.info(" (deint_show)");

        if(p->grow_mask > 0)
            grow_il(image, p->grow_mask);

        merge_plane(image->yy, p->frames[p->q_out].yy, image->w, image->h);
        merge_plane(image->cb, p->frames[p->q_out].cb, image->w, image->h);
        merge_plane(image->cr, p->frames[p->q_out].cr, image->w, image->h);
        }
    else if(p->min_ill > 0 && ill * ILL_SCALE < image->w * image->h * p->min_ill)
        m->vi3x.copy_image(image, p->frames + p->q_out);
    else
        {
        if(p->grow_mask > 0)
            grow_il(image, p->grow_mask);

        switch(p->mode)
            {
        case DEINT_SMART:
            if(p->debug)
                m->vi3x.info(" (deint_smart)");
            deint_smart(image->yy, p->frames[p->q_out].yy, image->w, image->h);
            deint_smart(image->cb, p->frames[p->q_out].cb, image->w, image->h);
            deint_smart(image->cr, p->frames[p->q_out].cr, image->w, image->h);
            break;
        case DEINT_SIMPLE:
            if(p->debug)
                m->vi3x.info(" (deint_simple)");
            deint_simple(image->yy, p->frames[p->q_out].yy, image->w, image->h);
            deint_simple(image->cb, p->frames[p->q_out].cb, image->w, image->h);
            deint_simple(image->cr, p->frames[p->q_out].cr, image->w, image->h);
            break;
        case DEINT_BLEND:
            if(p->debug)
                m->vi3x.info(" (deint_blend)");
            blend_plane(image->yy, p->frames[p->q_out].yy, image->w, image->h);
            blend_plane(image->cb, p->frames[p->q_out].cb, image->w, image->h);
            blend_plane(image->cr, p->frames[p->q_out].cr, image->w, image->h);
            break;
            }
        }
    image->il=p->frames[p->q_out].il;

    if(p->debug)
        m->vi3x.info("\n");

    p->q_out=(p->q_out + 1) % Q_DEPTH;
    --p->q_len;
    --p->q_fill;

    return m->out_image(m, image);
    }

static void mark_clear(vi3x_image_t *mark)
    {
    memset(mark->yy, 0, mark->img_sz);
    }

static void mark_all(vi3x_image_t *mark)
    {
    memset(mark->yy, MARK_YY, mark->plane_sz);
    memset(mark->cb, MARK_CB, mark->plane_sz);
    memset(mark->cr, MARK_CR, mark->plane_sz);
    }

typedef struct il_count_s
    {
    int d1, d2;
    int csign;
    } il_count_t;

static inline void add_il(il_count_t *ilc, int p0, int p1, int p2)
    {
    int v1, v2;

    v1=p0 - p1;
    v2=p2 - p1;
    ilc->csign+=v1 * v2;
    ilc->d1+=v1 * v1;
    ilc->d2+=v2 * v2;
    }

static int mark_il(vi3x_image_t *mark, vi3x_image_t const *img, int noise)
    {
    int end, xy;
    int ill=0;

    noise*=NOISE_SCALE;

    end=img->w * (img->h - 2);
    xy=0;
    while(xy < end)
        {
        int rend;
        for(rend=xy + img->w; xy < rend; ++xy)
            {
            il_count_t ilc={0, 0, 0};
            add_il(&ilc, img->yy[xy], img->yy[xy+img->w], img->yy[xy+img->w*2]);
            add_il(&ilc, img->cb[xy], img->cb[xy+img->w], img->cb[xy+img->w*2]);
            add_il(&ilc, img->cr[xy], img->cr[xy+img->w], img->cr[xy+img->w*2]);
            if(ilc.csign > CSIGN_THRESHOLD && ilc.d1 > noise && ilc.d2 > noise)
                {
                mark->yy[xy]=MARK_YY;
                mark->yy[xy+img->w]=MARK_YY;
                mark->cb[xy]=MARK_CB;
                mark->cb[xy+img->w]=MARK_CB;
                mark->cr[xy]=MARK_CR;
                mark->cr[xy+img->w]=MARK_CR;
                ill+=2;
                }
            }
        xy+=img->w;
        }

    for(end=xy + img->w; xy < end; ++xy)
        {
        il_count_t ilc={0, 0, 0};
        add_il(&ilc, img->yy[xy-img->w], img->yy[xy], img->yy[xy+img->w]);
        add_il(&ilc, img->cb[xy-img->w], img->cb[xy], img->cb[xy+img->w]);
        add_il(&ilc, img->cr[xy-img->w], img->cr[xy], img->cr[xy+img->w]);
        if(ilc.csign > 0 && ilc.d1 > noise && ilc.d2 > noise)
            {
            mark->yy[xy]=MARK_YY;
            mark->yy[xy+img->w]=MARK_YY;
            mark->cb[xy]=MARK_CB;
            mark->cb[xy+img->w]=MARK_CB;
            mark->cr[xy]=MARK_CR;
            mark->cr[xy+img->w]=MARK_CR;
            ill+=2;
            }
        }

    return ill;
    }

static void grow_il(vi3x_image_t *mark, int grow)
    {
    int rend;

    for(rend=mark->w; rend <= mark->count; rend+=mark->w)
        {
        int i;
        for(i=0; i < grow; ++i)
            {
            int xy=rend - mark->w;
            vi3x_sample_t p0=mark->yy[xy];
            for(++xy; xy < rend; ++xy)
                {
                if(mark->yy[xy] != p0)
                    {
                    p0=mark->yy[xy];
                    if(0 != p0)
                        {
                        mark->yy[xy-1]=MARK_YY;
                        mark->cb[xy-1]=MARK_CB;
                        mark->cr[xy-1]=MARK_CR;
                        }
                    else
                        {
                        mark->yy[xy]=MARK_YY;
                        mark->cb[xy]=MARK_CB;
                        mark->cr[xy]=MARK_CR;
                        }
                    }
                }
            }
        }
    }

static inline int sqd3(int p0, int p1, int p2)
    {
    int m=(p0 + p1 + p2)/3;

    p0-=m;
    p1-=m;
    p2-=m;

    return p0 * p0 + p1 * p1 + p2 * p2;
    }

static int unmark_static(
    vi3x_image_t *mark, vi3x_image_t const frames[Q_DEPTH], int motion, int ill)
    {
    int w=frames[0].w;
    int end, xy;

    end=w * (frames[0].h - 2);
    xy=0;
    while(xy < end)
        {
        int rend;
        for(rend=xy + w; xy < rend; ++xy)
            {
            if(mark->yy[xy])
                {
                int e=0;
                int row;
                for(row=0; row < w * 3; row+=w)
                    {
                    e+=sqd3(frames[0].yy[xy+row], frames[1].yy[xy+row],
                        frames[2].yy[xy+row]);
                    e+=sqd3(frames[0].cb[xy+row], frames[1].cb[xy+row],
                        frames[2].cb[xy+row]);
                    e+=sqd3(frames[0].cr[xy+row], frames[1].cr[xy+row],
                        frames[2].cr[xy+row]);
                    }
                e=e / (3 * 3 * 3); /* Normalize to 0-255**2 */
                if(e < motion)
                    {
                    mark->yy[xy]=0;
                    mark->yy[xy+w]=0;
                    mark->cb[xy]=0;
                    mark->cb[xy+w]=0;
                    mark->cr[xy]=0;
                    mark->cr[xy+w]=0;
                    ill-=2;
                    }
                }
            }
        xy+=w;
        }

    for(end=xy + w; xy < end; ++xy)
        {
        if(mark->yy[xy])
            {
            int e=0;
            int row;
            for(row=-w; row < w * 2; row+=w)
                {
                e+=sqd3(frames[0].yy[xy+row], frames[1].yy[xy+row],
                    frames[2].yy[xy+row]);
                e+=sqd3(frames[0].cb[xy+row], frames[1].cb[xy+row],
                    frames[2].cb[xy+row]);
                e+=sqd3(frames[0].cr[xy+row], frames[1].cr[xy+row],
                    frames[2].cr[xy+row]);
                }
            e= e / (3 * 3 * 3); /* Normalize to 0-255**2 */
            if(e < motion)
                {
                mark->yy[xy]=0;
                mark->yy[xy+w]=0;
                mark->cb[xy]=0;
                mark->cb[xy+w]=0;
                mark->cr[xy]=0;
                mark->cr[xy+w]=0;
                ill-=2;
                }
            }
        }

    return ill;
    }

static void merge_plane(
    vi3x_sample_t *dst, vi3x_sample_t const *src, int w, int h)
    {
    vi3x_sample_t *end;

    for(end=dst + w * h; dst < end; ++dst)
        {
        if(!*dst)
            *dst=*src;
        ++src;
        }
    }

static void deint_simple(
    vi3x_sample_t *dst, vi3x_sample_t const *src, int w, int h)
    {
    int x, y;

    for(y= h / 2 - 1; y > 0; --y)
        {
        memcpy(dst, src, w * sizeof(*dst));
        src+=w;
        dst+=w;
        for(x=w; x > 0; --x)
            {
            if(*dst)
                *dst=average(src[-w], src[w]);
            else
                *dst=*src;
            ++src;
            ++dst;
            }
        }

    memcpy(dst, src, w*sizeof(*dst));
    src+=w;
    dst+=w;
    for(x=w; x > 0; --x)
        {
        if(*dst)
            *dst=src[-w];
        else
            *dst=*src;
        ++src;
        ++dst;
        }
    }

static void deint_smart(
    vi3x_sample_t *dst, vi3x_sample_t const *src, int w, int h)
    {
    int x, y;

    for(y= h / 2 - 1; y > 0; --y)
        {
        memcpy(dst, src, w * sizeof(*dst));
        src+=w;
        dst+=w;
        for(x=SEARCH_WIDTH; x > 0; --x)
            {
            if(*dst)
                *dst=average(src[-w], src[w]);
            else
                *dst=*src;
            ++src;
            ++dst;
            }
        for(x=w - SEARCH_WIDTH * 2; x > 0; --x)
            {
            if(*dst)
                *dst=smart_avg(src - w, src + w);
            else
                *dst=*src;
            ++src;
            ++dst;
            }
        for(x=SEARCH_WIDTH; x > 0; --x)
            {
            if(*dst)
                *dst=average(src[-w], src[w]);
            else
                *dst=*src;
            ++src;
            ++dst;
            }
        }

    memcpy(dst, src, w*sizeof(*dst));
    src+=w;
    dst+=w;
    for(x=w; x > 0; --x)
        {
        if(*dst)
            *dst=src[-w];
        else
            *dst=*src;
        ++src;
        ++dst;
        }
    }

static inline int average(int a, int b)
    {
    return (a + b + 1) / 2;
    }

static int smart_avg(
    vi3x_sample_t const *px_u, vi3x_sample_t const *px_l)
    {
    int avg, delta;
    int delta_min, x_min;
    int x;

    avg=average(px_u[0], px_l[0]);
    delta=abs_diff(px_u[0], px_l[0]);

    if(delta <= MAX_DELTA)
        return avg;

    /* Go to the left */
    x_min=0;
    delta_min=delta;
    for(x=-2; x >= -SEARCH_WIDTH; --x)
        {
        int d=abs_diff(px_u[0], px_l[x]);
        if(delta_min > d)
            {
            delta_min=d;
            x_min=x;
            }
        }
    x_min=x_min / 2;
    if(0 != x_min)
        {
        delta_min=abs_diff(px_u[-x_min], px_l[x_min]);
        if(delta_min < delta)
            {
            int lavg=average(px_u[-x_min], px_l[x_min]);
            if((((int)(px_u[0]) - lavg) ^ ((int)(px_l[0]) - lavg)) < 0)
                {
                avg=lavg;
                delta=delta_min;
                }
            }
        }

    /* Go to the right */
    x_min=0;
    delta_min=delta;
    for(x=2; x <= SEARCH_WIDTH; ++x)
        {
        int d=abs_diff(px_u[0], px_l[x]);
        if(delta_min > d)
            {
            delta_min=d;
            x_min=x;
            }
        }
    x_min=x_min / 2;
    if(0 != x_min)
        {
        delta_min=abs_diff(px_u[-x_min], px_l[x_min]);
        if(delta_min < delta)
            {
            int lavg=average(px_u[-x_min], px_l[x_min]);
            if((((int)(px_u[0]) - lavg) ^ ((int)(px_l[0]) - lavg)) < 0)
                avg=lavg;
            }
        }

    return avg;
    }

static inline int abs_diff(int a, int b)
    {
    return abs(a - b);
    }

static void blend_plane(
    vi3x_sample_t *dst, vi3x_sample_t const *src, int w, int h)
    {
    vi3x_sample_t *end;

    /* First row */
    end=dst + w;
    while(dst < end)
        {
        if(*dst)
            *dst=average(src[0], src[w]);
        else
            *dst=*src;
        ++src;
        ++dst;
        }

    /* Middle rows */
    end+=w * (h - 2);
    while(dst < end)
        {
        if(*dst)
            *dst=average(src[0], average(src[-w], src[w]));
        else
            *dst=*src;
        ++src;
        ++dst;
        }

    /* Last row */
    end+=w;
    while(dst < end)
        {
        if(*dst)
            *dst=average(src[0], src[-w]);
        else
            *dst=*src;
        ++src;
        ++dst;
        }
    }
