/*
This file is part of video3x.

video3x is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

video3x is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with video3x.  If not, see <http://www.gnu.org/licenses/>.

Copyright 2011 Vadim Druzhin <cdslow@mail.ru>
*/

#include <string.h>
#include <limits.h>
#include <stdarg.h>
#include "vi3x_io.h"
#include "vi3x_utils.h"
#include "vi3x_help.h"

#define PARAM_TEXT_INDENT 4

static void help_module(
    char const *tag,
    char const *ver,
    char const *auth,
    char const *summary,
    char const *help_text,
    vi3x_arg_t const *args
    );
static void help_module_short(
    char const *tag,
    char const *ver,
    char const *auth,
    char const *summary,
    char const *help_text,
    vi3x_arg_t const *args
    );
static void help_args(vi3x_arg_t const *map);
static void help_param_start(void);
static void help_param_range(
    char const *name,
    char const *summary,
    char const *help_text,
    int def, int min, int max
    );
static void help_param_list(
    char const *name,
    char const *summary,
    char const *help_text,
    int def, int count, ...
    );
static void help_param_bool(
    char const *name,
    char const *summary,
    char const *help_text,
    int def,
    char const *false_text,
    char const *true_text
    );
static void help_module_end(void);
static void print_indent(char const *text, int indent);
static void help_none(void);
static void param_head(char const *name, char const *summary);
static void param_def(int def);

static const vi3x_help_t HelpFuncs=
    {
    vi3x_help,
    help_module,
    help_args,
    help_param_start,
    help_param_range,
    help_param_list,
    help_param_bool,
    help_module_end
    };

static const vi3x_help_t HelpShort=
    {
    (void (*)(char const *, ...))&help_none,
    help_module_short,
    (void (*)(vi3x_arg_t const *map))&help_none,
    &help_none,
    (void (*)(char const *, char const *, char const *, int, int, int))
        &help_none,
    (void (*)(char const *, char const *, char const *, int, int, ...))
        &help_none,
    (void (*)(char const *, char const *, char const *,
              int, char const *, char const *))&help_none,
    &help_none
    };

static void help_module(
    char const *tag,
    char const *ver,
    char const *auth,
    char const *summary,
    char const *help_text,
    vi3x_arg_t const *args
    )
    {
    vi3x_help("** ");
    help_module_short(tag, ver, auth, summary, help_text, args);
    vi3x_help("- Author: %s\n", auth);
    vi3x_help("- Version: %s\n", ver);
    print_indent(help_text, 4);
    if(NULL!=args)
        {
        help_param_start();
        help_args(args);
        }
    help_module_end();
    }

static void help_module_short(
    char const *tag,
    char const *ver,
    char const *auth,
    char const *summary,
    char const *help_text,
    vi3x_arg_t const *args
    )
    {
    (void)ver;       /* Unused */
    (void)auth;      /* Unused */
    (void)help_text; /* Unused */
    (void)args     ; /* Unused */
    vi3x_help("%s: %s\n", tag, summary); 
    }

static void help_args(vi3x_arg_t const *map)
    {
    vi3x_help_t hf;

    vi3x_help_funcs(&hf);

    while(NULL != map->name)
        {
        if(NULL != map->help)
            map->help(map, &hf);
        else
            {
            param_head(map->name, "-");
            param_def(map->def);
            }
        ++map;
        }
    }

static void help_param_start(void)
    {
    vi3x_help("  Parameters:\n");
    }

static void param_head(char const *name, char const *summary)
    {
    vi3x_help("  * %s: %s\n", name, summary);
    }

static void param_range(int min, int max)
    {
    vi3x_help("  - Value range: ");
    if(INT_MIN == min)
        vi3x_help("-oo");
    else
        vi3x_help("%d", min);
    vi3x_help(" ... ");
    if(INT_MAX == max)
        vi3x_help("oo");
    else
        vi3x_help("%d", max);
    vi3x_help("\n");
    }

static void param_list_va(int count, va_list ap)
    {
    vi3x_help("  - Values: ");
    while(count > 0)
        {
        vi3x_help("%d", va_arg(ap, int));
        (void)va_arg(ap, char *); /* Skip text */
        --count;
        if(count > 0)
            vi3x_help(", ");
        }
    vi3x_help("\n");
    }

static void param_def(int def)
    {
    vi3x_help("  - Default: %d\n", def);
    }

static void help_module_end(void)
    {
    vi3x_help("\n");
    }

static void help_param_range(
    char const *name,
    char const *summary,
    char const *help_text,
    int def, int min, int max
    )
    {
    param_head(name, summary);
    param_range(min, max);
    param_def(def);
    print_indent(help_text, PARAM_TEXT_INDENT);
    }

static void help_param_list(
    char const *name,
    char const *summary,
    char const *help_text,
    int def, int count, ...
    )
    {
    char indent[PARAM_TEXT_INDENT+1];
    va_list ap;
    int val;
    char *txt;

    memset(indent, ' ', PARAM_TEXT_INDENT);
    indent[PARAM_TEXT_INDENT]=0;

    param_head(name, summary);
    va_start(ap, count);
    param_list_va(count, ap);
    va_end(ap);
    param_def(def);
    print_indent(help_text, PARAM_TEXT_INDENT);
    va_start(ap, count);
    while(count > 0)
        {
        val=va_arg(ap, int);
        txt=va_arg(ap, char *);
        if(NULL != txt && 0 != *txt)
            {
            vi3x_help("%s%d %s", indent, val, txt);
            if('\n' != txt[strlen(txt)-1])
                vi3x_help("\n");
            }
        --count;
        }
    va_end(ap);
    }

static void help_param_bool(
    char const *name,
    char const *summary,
    char const *help_text,
    int def,
    char const *false_text,
    char const *true_text
    )
    {
    help_param_list(
        name, summary, help_text, def, 2, 0, false_text, 1, true_text);
    }

static void print_indent(char const *text, int indent)
    {
    char *buf, *s;

    buf=vi3x_alloc(strlen(text) + indent + 1, "print_indent");
    if(indent > 0)
        memset(buf, ' ', indent);

    s=buf + indent;
    while(0 != *text)
        {
        *s++=*text;
        if('\n' == *text)
            {
            *s=0;
            vi3x_help("%s", buf);
            s=buf + indent;
            }
        ++text;
        }

    if(s > buf + indent)
        {
        *s=0;
        vi3x_help("%s\n", buf);
        }

    vi3x_free(buf);
    }

void vi3x_help_funcs(vi3x_help_t *hf)
    {
    *hf=HelpFuncs;
    }

void vi3x_help_short(vi3x_help_t *hf)
    {
    *hf=HelpShort;
    }

static void help_none(void)
    {
    }
