/*
This file is part of video3x.

video3x is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

video3x is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with video3x.  If not, see <http://www.gnu.org/licenses/>.

Copyright 2011 Vadim Druzhin <cdslow@mail.ru>
*/

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include "vi3x_io.h"

static FILE *Log=NULL;

static int msg_va_log(char const *fmt, va_list ap);
static void msg_va_std(char const *fmt, va_list ap);
static void msg_va_err(char const *fmt, va_list ap);
static void msg_va(char const *fmt, va_list ap);
static void msg_va_help(char const *fmt, va_list ap);

#if defined(_WIN32) || defined(_WIN64)

#include <io.h>
#include <fcntl.h>

int vi3x_read(void *buf, int n)
    {
    static int is_bin=0;

    if(!is_bin)
        {
        _setmode(0, O_BINARY);
        is_bin=1;
        }

    return fread(buf, 1, n, stdin);
    }

int vi3x_write(void const *buf, int n)
    {
    static int is_bin=0;

    if(!is_bin)
        {
        _setmode(1, O_BINARY);
        is_bin=1;
        }

    return fwrite(buf, 1, n, stdout);
    }

#else

int vi3x_read(void *buf, int n)
    {
    return fread(buf, 1, n, stdin);
    }

int vi3x_write(void const *buf, int n)
    {
    return fwrite(buf, 1, n, stdout);
    }

#endif /* defined(_WIN32) || defined(_WIN64) */

void vi3x_die_oom(char const *tag)
    {
    fputs("\n", stderr);
    if(NULL != tag)
        {
        fputs("[", stderr);
        fputs(tag, stderr);
        fputs("] ", stderr);
        }
    fputs("Out of memory!\n", stderr);
    if(NULL != Log)
        {
        fputs("\n", Log);
        if(NULL != tag)
            {
            fputs("[", Log);
            fputs(tag, Log);
            fputs("] ", Log);
            }
        fputs("Out of memory!\n", Log);
        fclose(Log);
        }
    exit(EXIT_FAILURE);
    }

int vi3x_set_log(char const *fname)
    {
    FILE *l;

    if(NULL != Log)
        {
        fclose(Log);
        Log=NULL;
        }
    if(NULL != fname)
        {
        l=fopen(fname, "w");
        if(NULL != l)
            Log=l;
        else
            {
            vi3x_err("[vi3x_set_log] Error: Can't create log file '%s'!\n", fname);
            return 0;
            }
        }

    return 1;
    }

void vi3x_err(char const *fmt, ...)
    {
    va_list ap;

    va_start(ap, fmt);
    msg_va_log(fmt, ap);
    msg_va_err(fmt, ap);
    va_end(ap);
    }

void vi3x_warn(char const *fmt, ...)
    {
    va_list ap;

    va_start(ap, fmt);
    msg_va(fmt, ap);
    va_end(ap);
    }

void vi3x_info(char const *fmt, ...)
    {
    va_list ap;

    va_start(ap, fmt);
    msg_va(fmt, ap);
    va_end(ap);
    }

void vi3x_help(char const *fmt, ...)
    {
    va_list ap;

    va_start(ap, fmt);
    msg_va_help(fmt, ap);
    va_end(ap);
    }

void vi3x_info_va(char const *fmt, va_list ap)
    {
    msg_va(fmt, ap);
    }

static void msg_va(char const *fmt, va_list ap)
    {
    if(!msg_va_log(fmt, ap))
        msg_va_err(fmt, ap);
    }

static void msg_va_help(char const *fmt, va_list ap)
    {
    if(!msg_va_log(fmt, ap))
        msg_va_std(fmt, ap);
    }

static int msg_va_log(char const *fmt, va_list ap)
    {
    if(NULL != Log)
        {
        vfprintf(Log, fmt, ap);
        fflush(Log);
        return 1;
        }

    return 0;
    }

static void msg_va_err(char const *fmt, va_list ap)
    {
    vfprintf(stderr, fmt, ap);
    }

static void msg_va_std(char const *fmt, va_list ap)
    {
    vfprintf(stdout, fmt, ap);
    }
