/*
This file is part of video3x.

video3x is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

video3x is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with video3x.  If not, see <http://www.gnu.org/licenses/>.

Copyright 2011 Vadim Druzhin <cdslow@mail.ru>
*/

#include <string.h>
#include <stdint.h>
#include "vi3x_utils.h"
#include "vi3x_modules.h"
#include "vi3x_io.h"
#include "vi3x_image.h"
#include "vi3x_worker.h"

static int vi3x_fini(vi3x_workflow_t *wf);
static int vi3x_flush(vi3x_workflow_t *wf, vi3x_image_t *image);
static int vi3x_out_image(vi3x_module_t *m, vi3x_image_t *image);

vi3x_workflow_t *vi3x_create(void)
    {
    vi3x_workflow_t *wf;

    wf=vi3x_alloc(sizeof(*wf), "vi3x_create");
    wf->filters=NULL;
    wf->last_filter=NULL;

    return wf;
    }

void vi3x_destroy(vi3x_workflow_t *wf)
    {
    vi3x_module_t *m;

    while(NULL != wf->filters)
        {
        m=wf->filters->next;
        vi3x_free_module(wf->filters);
        wf->filters=m;
        }

    vi3x_free(wf);
    }

int vi3x_add_import(vi3x_workflow_t *wf, char const *args)
    {
    vi3x_module_t *m;

    m=vi3x_create_import(args);
    if(NULL == m)
        return 0;
    m->out_image=vi3x_out_image;

    if(NULL == wf->filters)
        {
        /* First module */
        wf->filters=m;
        wf->last_filter=m;
        }
    else if(vi3x_is_import(wf->filters))
        {
        /* Replace first module */
        m->next=wf->filters->next;
        *wf->filters=*m;
        vi3x_free_module(m);
        }
    else
        {
        /* Insert module */
        m->next=wf->filters;
        wf->filters=m;
        }

    return 1;
    }

int vi3x_add_export(vi3x_workflow_t *wf, char const *args)
    {
    vi3x_module_t *m;

    m=vi3x_create_export(args);
    if(NULL == m)
        return 0;

    if(NULL == wf->filters)
        {
        /* First module */
        wf->filters=m;
        wf->last_filter=m;
        }
    else if(vi3x_is_export(wf->last_filter))
        {
        /* Replace last module */
        *wf->last_filter=*m;
        vi3x_free_module(m);
        }
    else
        {
        /* Append module */
        wf->last_filter->next=m;
        wf->last_filter=m;
        }

    return 1;
    }

int vi3x_add_filter(vi3x_workflow_t *wf, char const *args)
    {
    vi3x_module_t *m;
    vi3x_module_t tmp;

    m=vi3x_create_filter(args);
    if(NULL == m)
        return 0;
    m->out_image=vi3x_out_image;

    if(NULL == wf->filters)
        {
        /* First module */
        wf->filters=m;
        wf->last_filter=m;
        }
    else if(vi3x_is_export(wf->last_filter))
        {
        /* Prepend last module */
        tmp=*wf->last_filter;
        *wf->last_filter=*m;
        *m=tmp;
        wf->last_filter->next=m;
        wf->last_filter=m;
        }
    else
        {
        /* Append module */
        wf->last_filter->next=m;
        wf->last_filter=m;
        }

    return 1;
    }

int vi3x_init(vi3x_workflow_t *wf)
    {
    vi3x_module_t *m;
    int fps, fps_div;

    /* Can't work without import */
    if(!vi3x_is_import(wf->filters))
        {
        vi3x_err("[vi3x_init] Error: Missing import module!\n");
        return 0;
        }
    /* Can't work without export */
    if(!vi3x_is_export(wf->last_filter))
        {
        vi3x_err("[vi3x_init] Error: Missing export module!\n");
        return 0;
        }

    m=wf->filters;
    m->w=-1;
    m->h=-1;
    m->fps=25;
    m->fps_div=1;
    for(;;)
        {
        fps=m->fps;
        fps_div=m->fps_div;
        if(!m->init(m))
            break;
        if(NULL == m->process_image)
            {
            vi3x_err("[vi3x_init] Error: Module '%s' did not provide "
                "'process_image'!\n", m->name);
            break;
            }
        if((fps != m->fps || fps_div != m->fps_div) && *m->range > 0)
            {
            vi3x_err("[vi3x_init] Error: Module '%s' changes FPS, "
                "so can't have ranges!\n", m->name);
            break;
            }
        if(NULL == m->next)
            return 1;
        m->next->w=m->w;
        m->next->h=m->h;
        m->next->fps=m->fps;
        m->next->fps_div=m->fps_div;
        m=m->next;
        }

    vi3x_err("[vi3x_init] Error: Init failed on '%s'!\n", m->name);

    /* Cleanup failed initialization */
    vi3x_fini(wf);

    return 0;
    }

int vi3x_run(vi3x_workflow_t *wf)
    {
    vi3x_image_t image;
    vi3x_sample_t *buf;
    int ret;

    vi3x_init_image(&image, wf->filters->w, wf->filters->h, NULL);
    buf=vi3x_alloc(image.img_sz, "vi3x_run");
    vi3x_init_image(&image, wf->filters->w, wf->filters->h, buf);

    /* Process frames */
    do
        {
        ret=wf->filters->process_image(wf->filters, &image);
        }
    while(ret != VI3X_RET_FINISH && ret != VI3X_RET_ERR);

    ret=vi3x_flush(wf, &image) && ret;
    ret=vi3x_fini(wf) && ret;

    vi3x_free(buf);

    return ret != VI3X_RET_ERR;
    }

static int vi3x_fini(vi3x_workflow_t *wf)
    {
    vi3x_module_t *m;
    int ret=VI3X_RET_OK;

    for(m=wf->filters; NULL != m; m=m->next)
        {
        if(NULL != m->fini)
            ret=m->fini(m) && ret;
        }

    return ret;
    }

static int vi3x_flush(vi3x_workflow_t *wf, vi3x_image_t *image)
    {
    vi3x_module_t *m;
    int ret=VI3X_RET_OK;

    for(m=wf->filters; NULL != m; m=m->next)
        {
        if(!m->disabled && NULL != m->flush)
            ret=m->flush(m, image) && ret;
        }

    return ret;
    }

static int vi3x_out_image(vi3x_module_t *m, vi3x_image_t *image)
    {
    if(NULL == m->next || NULL == m->next->process_image)
        {
        vi3x_err("[vi3x_out_image] Error: Broken filter chain after '%s'!\n", m->name);
        return 0;
        }

    m=m->next;
    ++m->frame_n;

    if(m->frame_n == *m->range_ptr)
        {
        /* First or last frame in range */
        ++m->range_ptr;
        /* Ranges are inclusive, so process frame in any case */
        if(!m->process_image(m, image))
            return 0;
        /* Enable or disable module */
        m->disabled=!m->disabled;
        if(!m->disabled && m->frame_n == *m->range_ptr)
            {
            /* It was single frame range */
            ++m->range_ptr;
            /* Disable module again */
            m->disabled=!m->disabled;
            }
        if(m->disabled)
            {
            /* It was last frame in range */
            if(NULL != m->flush)
                return m->flush(m, image);
            }
        return 1;
        }

    if(m->disabled)
        /* Skip frame */
        return m->out_image(m, image);
    else
        /* Process frame */
        return m->process_image(m, image);
    }
